import argparse
import os
import numpy as np
import scanpy as sc
import muon as mu
import matplotlib.pyplot as plt
import panpipes.funcs as pnp

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()
parser.add_argument("--filtered_mudata",
                    default=None,
                    help="")  
parser.add_argument("--raw_mudata",
                    default=None,
                    help="")
parser.add_argument("--channel_col",
                    default=None,
                    help="")
parser.add_argument("--normalisation_methods", default="clr,dsb",
                   help="comma separated list of normalisation methods")
parser.add_argument("--clr_margin",
                    default=0,
                    help="")    
parser.add_argument("--quantile_clipping",
                    default=False,
                    help="")    
parser.add_argument("--figpath",
                    default="./prot_figures",
                    help="")  
parser.add_argument("--save_mtx",
                    default=False,
                    help="")  
parser.add_argument("--save_mudata_path",
                    default=None,
                    help="")   
              

args, opt = parser.parse_known_args()
save_mtx=pnp.pp.check_for_bool(args.save_mtx)

norm_methods = args.normalisation_methods.split(',')

L.info(args)
# load filtered data - if use_umon is True this will return a mudata object, else returns an mudata object

L.info("reading filtered mudata object")
try:
    all_mdata = mu.read(args.filtered_mudata)
except FileNotFoundError:
    sys.exit("filtered_mudata file not found")

# load raw data
if 'dsb' in norm_methods:
    if args.raw_mudata is not None:
        L.info("reading raw file object")
        try:
            all_mdata_raw = mu.read(args.raw_mudata)
            # subset to just the channel 
        except FileNotFoundError:
            sys.exit("raw_mudata object not found")
    else:
        sys.exit("must specify a raw mudata to run dsb, containing both rna and prot")


# find isotypes columns 
if "isotype" in all_mdata["prot"].var.columns:
    isotypes = list(all_mdata["prot"].var_names[all_mdata["prot"].var.isotype])
else:
    isotypes = None
    
L.info("isotypes found:")
L.info(isotypes)


# save raw counts
all_mdata["prot"].layers["raw_counts"] = all_mdata["prot"].X.copy()


if args.channel_col is not None:
    ## running per channel (parallelise in future)
    channel_cols = list(pnp.processing.mu_get_obs( all_mdata, 
        features=[args.channel_col],
        modalities=["prot"]).iloc[:,0].unique())
    mdata_per_sample = {si: all_mdata[all_mdata.obs[args.channel_col] == si] for si in channel_cols}
    # this data is only necessary if running dsb
    if 'dsb' in norm_methods:
        mdata_raw_per_sample = {si: all_mdata_raw[all_mdata_raw.obs[args.channel_col] == si] for si in channel_cols}
    for si in channel_cols:
        # si = list(mdata_per_sample.keys())[0]
        L.info(si)
        mdata=mdata_per_sample[si]
        plot_features = list(mdata["prot"].var_names)
        plot_features.sort()
        plot_features = [x for x in plot_features if x not in isotypes]
        # first run clr
        if 'clr' in norm_methods:
            # make sure to start from raw counts
            mdata["prot"].X = mdata["prot"].layers["raw_counts"]
            pnp.scmethods.run_prot_normalise(mdata=mdata, mdata_raw=None,
                    method="clr",
                    clr_margin=int(args.clr_margin))
            L.info("saving ridgeplot")
            pnp.plotting.ridgeplot(mdata["prot"], features=plot_features, layer="clr",  splitplot=6)
            plt.savefig(os.path.join(args.figpath, si + "_clr_ridgeplot.png"))
            if isotypes is not None:
                pnp.plotting.ridgeplot(mdata["prot"], features=isotypes, layer="clr",  splitplot=1)
                plt.savefig(os.path.join(args.figpath, si + "_clr_ridgeplot_isotypes.png"))
            # save out the data in what format?
            if save_mtx:
                pnp.io.write_10x_counts(mdata["prot"], os.path.join("prot_clr" , si ), layer="raw_counts")
        # then run dsb
        if 'dsb' in norm_methods:
            mdata_raw = mdata_raw_per_sample[si]
            mdata = mdata_per_sample[si]
            # make sure to start from raw counts
            mdata["prot"].X = mdata["prot"].layers["raw_counts"]
            # plot hiustogram of bg
            mdata_raw["rna"].obs["log10umi"] = np.array(np.log10(mdata_raw["rna"].X.sum(axis=1) + 1)).reshape(-1)
            # mu.pl.histogram(mdata_raw["rna"], ["log10umi"], bins=50)
            # plt.savefig(os.path.join(args.figpath, si + "_log10umi.png"))
            # run dsb
            pnp.scmethods.run_prot_normalise(mdata=mdata, 
                                            mdata_raw=mdata_raw,
                                            method="dsb", 
                                            isotypes=isotypes)
            pnp.plotting.ridgeplot(mdata["prot"], features=plot_features, layer="dsb",  splitplot=6)
            plt.savefig(os.path.join(args.figpath, si + "_dsb_ridgeplot.png"))
            if isotypes is not None:
                pnp.plotting.ridgeplot(mdata["prot"], features=isotypes, layer="dsb",  splitplot=1)
                plt.savefig(os.path.join(args.figpath, si + "_dsb_ridgeplot_isotypes.png"))
            if save_mtx:
                pnp.io.write_10x_counts(mdata["prot"], os.path.join("prot_dsb" , si), layer="raw_counts")
else:
    # run on all the data (not on a channel basis)
    # first run clr
     # do the ridgeplots for clr
    plot_features = list(all_mdata["prot"].var_names)
    plot_features.sort()
    plot_features = [x for x in plot_features if x not in isotypes]
    if 'clr' in norm_methods:
        # make sure to start from raw counts
        all_mdata["prot"].X = all_mdata["prot"].layers["raw_counts"]
        # run normalise
        pnp.scmethods.run_prot_normalise(mdata=all_mdata, 
                mdata_raw=None,
                method="clr",
                clr_margin=int(args.clr_margin))
        pnp.plotting.ridgeplot(all_mdata["prot"], features=plot_features, layer="clr",  splitplot=6)
        plt.savefig(os.path.join(args.figpath, "clr_ridgeplot.png"))
        if isotypes is not None:
            pnp.plotting.ridgeplot(all_mdata["prot"], features=isotypes, layer="clr",  splitplot=1)
            plt.savefig(os.path.join(args.figpath, "clr_ridgeplot_isotypes.png"))
        # save out the data in what format?
        if save_mtx:
            pnp.io.write_10x_counts(all_mdata["prot"], os.path.join("prot_clr"), layer="clr")
    if 'dsb' in norm_methods:
        # make sure to start from raw counts
        all_mdata["prot"].X = all_mdata["prot"].layers["raw_counts"]
        # then run dsb
        # comput log 10 umi
        all_mdata_raw["rna"].obs["log10umi"] = np.array(np.log10(all_mdata_raw["rna"].X.sum(axis=1) + 1)).reshape(-1)
        mu.pl.histogram(all_mdata_raw["rna"], ["log10umi"], bins=50)
        plt.savefig(os.path.join(args.figpath, "all_log10umi.png"))
        pnp.scmethods.run_prot_normalise(mdata=all_mdata, 
                mdata_raw=all_mdata_raw, 
                method="dsb",
                isotypes=isotypes) 
        # apply quantile clipping # discussed in FAQs https://cran.r-project.org/web/packages/dsb/vignettes/dsb_normalizing_CITEseq_data.html
        if args.quantile_clipping:
            all_mdata['prot'].layers['dsb_clipped'] = pnp.scmethods.quantile_clipping(all_mdata['prot'], layer="dsb", inplace=False)
            all_mdata['prot'].X = all_mdata['prot'].layers['dsb_clipped']
        # plot ridgeplot
        pnp.plotting.ridgeplot(all_mdata["prot"], features=plot_features, layer="dsb",  splitplot=6)
        plt.savefig(os.path.join(args.figpath, "dsb_ridgeplot.png"))
        if isotypes is not None:
            pnp.plotting.ridgeplot(all_mdata["prot"], features=isotypes, layer="dsb",  splitplot=1)
            plt.savefig(os.path.join(args.figpath, "dsb_ridgeplot_isotypes.png"))
        # save out the data in what format?
        if save_mtx:
            pnp.io.write_10x_counts(all_mdata["prot"], os.path.join("prot_dsb"), layer="dsb")

    # run pca on dsb normalised (if dsb was run otherwise clr is in X)
    sc.tl.pca(all_mdata['prot'], n_comps=50, svd_solver='arpack', random_state=0) 

    if args.save_mudata_path is not None:
        all_mdata.update()
        pnp.io.write_anndata(all_mdata, args.save_mudata_path, use_muon=True, modality='all')



L.info("Done")




import multiprocessing 

import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import argparse
import os
import gc
import muon as mu
from cgatcore import pipeline as P
import panpipes.funcs as pp
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice, X_is_raw

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--scaled_anndata',
                    default='adata_scaled.h5ad',
                    help='a preprocessed mudata/anndata')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_MultiVI.csv',
                    help='')
parser.add_argument('--use_gpu', default=False,
                    help='')
parser.add_argument('--integration_col_categorical', default='batch',
                    help='')
parser.add_argument('--n_factors', default="",
                    help='number of factors to train the model with - optional')
parser.add_argument('--n_iterations', default='',
                    help='upper limit on the number of iterations')
parser.add_argument('--convergence_mode', default='fast',
                    help='fast, medium or slow')
parser.add_argument('--save_parameters', default=bool,
                    help='whether to save parameters models')
parser.add_argument('--outfile_model', default="",
                    help="if args.save_parameters true, path to HDF5 file to store the model")
parser.add_argument('--figdir', default='./figures',
                    help='')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs", default=50)
parser.add_argument('--neighbors_method', default="scanpy",
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k', default=30,
                    help="neighbors k")
parser.add_argument('--neighbors_metric',default="euclidean",
                    help="neighbor metric, e.g. euclidean or cosine")

args, opt = parser.parse_known_args()

L.info(args)
# scanpy settings
sc.set_figure_params(facecolor="white")
sc.settings.autoshow = False
sc.settings.figdir = args.figdir


# load parameters

threads_available = multiprocessing.cpu_count()
params = pp.io.read_yaml("pipeline.yml")

mdata = mu.read(args.scaled_anndata)

if params['multimodal']['mofa']['modalities'] is not None:
    modalities= params['multimodal']['mofa']['modalities']
    modalities = [x.strip() for x in modalities.split(",")]
    L.info(f"using modalities :{modalities}")
    removed_mods = None
    if all(x in modalities for x in mdata.mod.keys()):
        tmp = mdata.copy()
        L.info('using all modalities')
    else:
        tmp = mdata.copy()
        removed_mods = list(set(mdata.mod.keys()) - set(modalities))
        L.info(f"removing modalities {removed_mods}")
        for rmod in removed_mods:
            del tmp.mod[rmod]
else:
    L.warning("""you specified no modalities, so i will default to all available
                this may be a problem if you have repertoire in here""")
    removed_mods = None
    tmp = mdata.copy()  

print(tmp.mod.keys())

mu.pp.intersect_obs(tmp)

mofa_kwargs={}
#expected args:
# n_factors: 10
# n_iterations: 1000
# convergence_mode: fast, medium, slow
# save_parameters: False
# #if save_parameters True, set the following, otherwise leave blank
# outfile_model:

mofa_kwargs = params['multimodal']['mofa']
del mofa_kwargs['modalities']

if mofa_kwargs['filter_by_hvg']:
    mofa_kwargs['use_var'] = "highly_variable"
    del mofa_kwargs['filter_by_hvg']
    for mod in tmp.mod.keys():
        if "highly_variable" not in tmp[mod].var.columns:
            print(mod)
            tmp[mod].var["highly_variable"] = True
        
    tmp.update()
    


if args.integration_col_categorical is not None:
    if args.integration_col_categorical in tmp.obs.columns:
        mofa_kwargs['groups_label'] = args.integration_col_categorical

# default is to read yaml and parse directly to kwargs.
# if the defaults expected params are parsed by the script in some other way
# they will overwrite the initial reading of the yml

if mofa_kwargs['save_parameters'] is None:
    if args.save_parameters is not None:
        mofa_kwargs['save_parameters'] = check_for_bool(args.save_parameters)
        if args.outfile_model is not None:
            mofa_kwargs['outfile'] = args.outfile_model

if mofa_kwargs['n_factors'] is None:
    if args.n_factors is not None:
        mofa_kwargs['n_factors'] = int(args.n_factors)

if mofa_kwargs['n_iterations'] is None:
    if args.n_iterations is not None:
        mofa_kwargs['n_iterations'] = int(args.n_iterations)

if mofa_kwargs['convergence_mode'] is None:
    if args.convergence_mode is not None:
        mofa_kwargs['convergence_mode'] = args.convergence_mode
        
# are we using the gpu?
if args.use_gpu is True or args.use_gpu=='True':
     mofa_kwargs['gpu_mode'] = True

L.info(mofa_kwargs)
L.info(tmp.var.columns)


mu.tl.mofa(tmp, **mofa_kwargs)
#This adds X_mofa embeddings to the .obsm slot of the MuData object
#write the discovered latent rep to the original mudata object
mdata.obsm['X_mofa'] = tmp.obsm['X_mofa'].copy()   

if int(args.neighbors_n_pcs) > mdata.obsm['X_mofa'].shape[1]:
    L.warning(f"N PCs is larger than X_mofa dimensions, reducing n PCs to  {mdata.obsm['X_mofa'].shape[1]-1}")
n_pcs= min(int(args.neighbors_n_pcs), mdata.obsm['X_mofa'].shape[1]-1)


run_neighbors_method_choice(mdata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=n_pcs, #this should be the # rows of var, not obs ???????
    metric=args.neighbors_metric, 
    use_rep='X_mofa',
    nthreads=max([threads_available, 6]))

sc.tl.umap(mdata, min_dist=0.4)
sc.tl.leiden(mdata,  key_added="leiden_mofa")

umap = pd.DataFrame(mdata.obsm['X_umap'], mdata.obs.index)
umap.to_csv(args.output_csv)

if removed_mods is not None:
    for rmd in removed_mods:
        tmp.mod[rmd] = mdata.mod[rmd].copy()

tmp.write("tmp/mofa_scaled_adata.h5mu")


L.info("Done")




#!/bin/py
# aggregate_cellranger.py

# takes cellranger multi outputs and cellranger count outputs, and makes them consistent for plotting.


import pandas as pd
import glob
import os
import re
import argparse
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()
parser.add_argument('--pipe_df',
                    default='',
                    help='')
parser.add_argument('--output_file',
                    default='',
                    help='')
parser.add_argument('--figdir',
                    default='./figures',
                    help='')
parser.add_argument('--cellranger_column_conversion_df',
                    default='',
                    help='')
parser.set_defaults(verbose=True)
args = parser.parse_args()

L.info(args)

def check_path(path):
    if os.path.isfile(path):
        return os.path.dirname(path)
    elif os.path.isdir(path):
        return path
    else:
        return f"{path} does not exist or is an unrecognized type."

def get_metrics_summary_path(path,sample_id=None):
    """ infers the path to metrics_summary.csv based on what cellranger algorithm is used 
    and checks the file exists

    Args:
        path (str):path to the folder called 'outs'
        sample_id (str): Required if path is to cellranger multi outputs, defaults to None
    """    
    # subset path to only go up to 'outs'
    if 'outs' not in path:
        print("you are parsing a cellranger output but your path to raw data doesn't end wth the outs folder")
        path = check_path(path)
    else:
        path = path.split("outs")[0] + "outs"
    outpath=None
    # use the path to cellranger count or vdj outputs as default
    if os.path.exists(os.path.join(path, 'metrics_summary.csv') ):
        outpath = os.path.join(path, 'metrics_summary.csv') 
    elif os.path.exists(os.path.join(path, 'summary.csv') ):
        outpath = os.path.join(path, 'summary.csv') 
    elif sample_id is not None and os.path.exists(os.path.join(path, 'per_sample_outs', sample_id, 'metrics_summary.csv') ):
        outpath = os.path.join(path, 'per_sample_outs', sample_id, 'metrics_summary.csv')
    elif sample_id is None and os.path.exists(os.path.join(path, 'per_sample_outs')):
        print('input folder appears to be from cellranger multi but no sample_id is given')
    else:
        # use the alternative path from cellranger_multi outputs
        print('path not found')
    return outpath


def detect_cellranger_algorithm(pth):
    if 'per_sample_outs' in pth:
        filetype='multi'
    else:
        filetype='count'
    return filetype


def get_all_unique_paths(pipe_df):
    """subsets all columns of pd.Dataframe which are suffixed 'path' and stacks the result

    Args:
        pipe_df (pd.DataFrame): dataframne that contains at least one columns suffixed 'path'

    Returns:
        pd.DataFrame: stacked dataframe of paths
    """    
    pipe_df = pipe_df.set_index('sample_id')
    pipe_df = pipe_df.loc[:,pipe_df.columns.str.endswith('path')]
    all_paths = pipe_df.stack().reset_index()
    all_paths.columns = ['sample_id', 'path_type', 'path']
    all_paths = all_paths.drop_duplicates()
    # rename path_type to match cellranger terminolgy
    recode_dict = {'rna_path': "Gene Expression", 
                   'prot_path': "Antibody Capture", 
                   'tcr_path': 'VDJ T', 
                   'bcr_path': 'VDJ B',
                   'atac_path': 'ATAC' }
    all_paths['path_type'] = all_paths['path_type'].replace(recode_dict)
    return all_paths


def parse_10x_cellranger_multi(path_df,path_col='metrics_summary_path'):
    """takes the metrics_summary.csv from cellranger multi for each sample 
    and concantenates into on pd.DataFrame (long format)

    Args:
        path_df (pd.DataFrame): pandas dataframe with two columns sample_id and and 
                            path which contains the path to the outs folder.
        path_col (str): colname for the columns containing the paths

    Returns:
        pd.DataFrame: 7 columns ['sample_id', 'category', 'library_type', 'grouped_by', 
        'group_name', metric_name', 'metric_value']
    """   
    # subset to unique rows to save loading in duplicate data
    path_df = path_df[['sample_id', path_col]].drop_duplicates()
    # read and concat
    msums = pd.concat([pd.read_csv(f) for f in path_df[path_col]], 
                    keys=path_df['sample_id'], 
                    names=['sample_id']).reset_index().drop(columns='level_1')
    # remove spaces and capitils from columns names
    msums.columns = [re.sub(' ', '_', x.lower()) for x in msums.columns]
    # add a fix for version < cellranger 7
    ranger = {"library_or_sample":"category"}
    msums.rename(columns =ranger,inplace=True)
    # convert percentages from string to numeric
    msums['metric_value'] = [re.sub(",|%", "", x) for x in msums['metric_value']]
    msums['metric_value'] = msums['metric_value'].astype(float)
    # sort outputs
    msums = msums.sort_values(['metric_name', 'sample_id'])
    return msums


def parse_10x_cellranger_count(path_df, convert_df,  path_col='metrics_summary_path'):
    """takes the metrics_summary.csv from cellranger count output for each sample 
    and concantenates into on pd.DataFrame (long format)

    Args:
        path_df (pd.DataFrame): pandas dataframe with three columns sample_id, path 
                                path which contains the path to the outs folder.and 
                                path_type which contains the type of library (one of ['Gene Expression', 'Antibody Capture', 'VDJ T', 'VDJ B'])
        path_col (str): colname for the columns containing the paths

    Returns:
        pd.DataFrame: 7 columns ['sample_id', 'category', 'library_type', 'grouped_by', 
        'group_name', metric_name', 'metric_value']
    """   
    
    # read and concat
    msums = pd.concat([pd.read_csv(f) for f in path_df[path_col]], 
                    keys=[(x, y) for x, y in zip(path_df['sample_id'], 
                                                 path_df['path_type'])], #count
                    names=['sample_id','path_type'])
    msums = msums.stack().reset_index().drop(columns='level_2')
    # make the columns match the conversion df
    msums.columns = ['sample_id', 'library_type', 'count_metric_name', 'metric_value']
    # update the columns names
    msums = msums.merge(convert_df).drop(columns=['count_metric_name'])
    msums['group_name'] = msums['sample_id']
    msums = msums[['sample_id','category', 'library_type', 'grouped_by', 'group_name', 'metric_name', 'metric_value']]
    # convert percentages from string to numeric
    msums['metric_value'] = [re.sub(",|%", "", str(x)) for x in msums['metric_value']]
    msums['metric_value'] = msums['metric_value'].astype(float)
    msums = msums.sort_values(['library_type', 'metric_name', 'sample_id'])
    return msums
    

convert_df = pd.read_csv(args.cellranger_column_conversion_df, sep='\t')   

# get the paths
pipe_df = pd.read_csv(args.pipe_df, sep='\t')
L.info('finding all the cellranger paths')
all_paths_df = get_all_unique_paths(pipe_df)
all_paths_df['metrics_summary_path'] = all_paths_df.apply(lambda x: get_metrics_summary_path(path=x.path, sample_id=x.sample_id), axis=1)
# all_paths_df_uniq = all_paths_df.drop(columns=['path', 'path_type']).drop_duplicates().reset_index(drop=True)
all_paths_df['cellranger_type'] = [detect_cellranger_algorithm(x) for x in all_paths_df['metrics_summary_path']]
L.info('cellranger metrics_summary files found')
L.info(all_paths_df)

# parse and concatenate the tenx x metrics summary files into long format
tenx_metrics = []
if any(all_paths_df['cellranger_type']=='multi'):
    tenx_metrics.append(parse_10x_cellranger_multi(all_paths_df[all_paths_df['cellranger_type']=='multi']))

if any(all_paths_df['cellranger_type']=='count'):
    tenx_metrics.append(parse_10x_cellranger_count(all_paths_df[all_paths_df['cellranger_type']=='count'], convert_df))



tenx_metrics_full = pd.concat(tenx_metrics)
tenx_metrics_full = tenx_metrics_full.sort_values(['library_type', 'metric_name'])
tenx_metrics_full.to_csv(args.output_file, index=False)
L.info('done')


# tenx_metrics['metric_value'] = [re.sub(",|%", "", x) for x in tenx_metrics['metric_value']]
# tenx_metrics['metric_value'] = tenx_metrics['metric_value'].astype(float)
# split by library_type and metric_name 
for idx, row in tenx_metrics_full[['library_type','metric_name']].drop_duplicates().iterrows():
    mn = row['metric_name']
    lt = row['library_type']
    plt_df = tenx_metrics_full[(tenx_metrics_full['library_type'] == lt) & (tenx_metrics_full['metric_name'] == mn)]  
    if len(plt_df.category.unique()) > 1:
        # if multiple categories i.e. Cells and Libraries, these are likely duplicate rows
        plt_df = plt_df[['sample_id', 'metric_name', 'library_type' , 'metric_value']].drop_duplicates()
    if len(plt_df['sample_id']) == len(plt_df['sample_id'].unique()):
        fig = sns.barplot(plt_df, x='sample_id', y='metric_value', color='grey')
        fig.set_xticklabels(fig.get_xticklabels(), rotation=90)
        fig.set_title(lt + ':' + mn)
        plt.savefig(os.path.join(args.figdir, lt + '-' + mn + '.png'), bbox_inches='tight')
    else:
        # do a boxplot instead
        fig = sns.boxplot(plt_df, x='sample_id', y='metric_value', color='grey')
        fig.set_xticklabels(fig.get_xticklabels(), rotation=90)
        fig.set_title(lt + ':' + mn)
        plt.savefig(os.path.join(args.figdir, lt + '-' + mn + '.png'), bbox_inches='tight')
    plt.clf()


## gene expression  extra plots -----------------------------

rna_tenx_metrics = tenx_metrics_full[tenx_metrics_full['library_type'] == 'Gene Expression']

# sequencing_saturaion plot scatter plot
plot_metrics = ['Sequencing saturation', 'Mean reads per cell', 'Estimated number of cells', 'Number of reads', 'Median UMI counts per cell']
plt_df = rna_tenx_metrics[rna_tenx_metrics.metric_name.isin(plot_metrics)]
plt_df = plt_df[['sample_id', 'metric_name', 'metric_value']]
plt_tab = plt_df.pivot_table(index='sample_id', columns='metric_name', values='metric_value', aggfunc=sum)
# sns.scatterplot(data=plt_tab, 
#                 x='Sequencing saturation', 
#                 y='Number of reads', size='Mean reads per cell',sizes=(20, 200))

if 'Sequencing saturation' in plt_df.metric_name.unique(): 
    f, ax = plt.subplots()
    points = ax.scatter(x=plt_tab['Sequencing saturation'], 
                y=plt_tab['Number of reads'], 
                c=plt_tab['Estimated number of cells'])
    # define labels
    plt.xlabel('Sequencing saturation')
    plt.ylabel('Number of reads')
    # define color bar
    cbar = f.colorbar(points)
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel('Estimated # of cells', rotation=270)
    # put it on a grid
    plt.grid(axis='both', color='0.95')
    # save
    plt.savefig(os.path.join(args.figdir,'10x_sequencing_saturation_summary.png'))
    plt.clf()

# plot 2 -  cells vs UMI counts per cell

fig, ax = plt.subplots()
x=list(np.log10(plt_tab['Estimated number of cells']))
y=list(np.log10(plt_tab['Median UMI counts per cell']))
ax.scatter(x=x, 
            y=y)
# annotate each point
for i, txt in enumerate(plt_tab.index):
    ax.annotate(txt, (x[i], y[i]), size=8, ha='center')
plt.xlabel('log10(Estimated number of cells)')
plt.ylabel('log10(Median UMI counts per cell)')
# put a boundary box around the good cells
plt.vlines(x=(np.log10(500), np.log10(20000)), 
           ymin=min(1.5, min(y)),
           ymax=max(5, max(y)),
           colors='black', linestyles='dotted'
           )
plt.hlines(y=(3, 4), 
            xmin=min(1.5, min(x)),
           xmax=max(5, max(x)),
           colors='black', linestyles='dotted'
)

# put it on a grid
plt.grid(axis='both', color='0.95')
# save
plt.savefig(os.path.join(args.figdir,'10x_cells_by_UMIs.png'))
plt.clf()


from ruffus import *
import sys
import os
from cgatcore import pipeline as P
import glob


PARAMS = P.get_parameters(
    ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
     "pipeline.yml"])

PARAMS['py_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'python_scripts')
PARAMS['r_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'R_scripts')
PARAMS['resources_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")

job_kwargs = {}

if PARAMS['condaenv'] is not None:
    job_kwargs["job_condaenv"] = PARAMS['condaenv']



if (PARAMS["input_spatial"] is not None) and (not os.path.exists(PARAMS["input_spatial"])):
    sys.exit("Spatial input directory doesn't exist or cannot be found")



def gen_filter_jobs():
    input_paths_spatial=glob.glob(os.path.join(PARAMS["input_spatial"],"*.h5mu"))
    input_singlecell = PARAMS["input_singlecell"]
    for input_spatial in input_paths_spatial:
        sample_prefix = os.path.basename(input_spatial)
        sample_prefix = sample_prefix.replace(".h5mu","")
        outfile_spatial = "cell2location.output/" + sample_prefix + "/Cell2Loc_spatial_output.h5mu"
        yield input_spatial, outfile_spatial, sample_prefix, input_singlecell



@mkdir("logs")
@mkdir("figures")
@active_if(PARAMS['Cell2Location_run'] is True)
@mkdir("figures/Cell2Location")
@mkdir("cell2location.output")
@files(gen_filter_jobs)
def run_cell2location(input_spatial, outfile_spatial, sample_prefix, input_singlecell):

    figdir = "./figures/Cell2Location/" + sample_prefix
    output_dir = "./cell2location.output/" + sample_prefix
    log_file = "Cell2Location_" + sample_prefix + ".log"
    cmd = """
        python %(py_path)s/run_cell2location.py
        --input_spatial %(input_spatial)s
        --input_singlecell %(input_singlecell)s
        --figdir %(figdir)s
        --output_dir %(output_dir)s
              """
    
    # feature selection paramaters
    if PARAMS['Cell2Location_gene_list'] is not None:
        cmd += " --gene_list %(Cell2Location_gene_list)s"
    if PARAMS['Cell2Location_remove_mt'] is not None:
        cmd += " --remove_mt %(Cell2Location_remove_mt)s"
    if PARAMS['Cell2Location_cell_count_cutoff'] is not None:
        cmd += " --cell_count_cutoff %(Cell2Location_cell_count_cutoff)s"
    if PARAMS['Cell2Location_cell_percentage_cutoff2'] is not None:
        cmd += " --cell_percentage_cutoff2 %(Cell2Location_cell_percentage_cutoff2)s"
    if PARAMS['Cell2Location_nonz_mean_cutoff'] is not None:
        cmd += " --nonz_mean_cutoff %(Cell2Location_nonz_mean_cutoff)s"
    
    # parameters for the reference model
    if PARAMS['Cell2Location_labels_key_reference'] is not None:
        cmd += " --labels_key_reference %(Cell2Location_labels_key_reference)s"
    if PARAMS['Cell2Location_batch_key_reference'] is not None:
        cmd += " --batch_key_reference %(Cell2Location_batch_key_reference)s"
    if PARAMS['Cell2Location_layer_reference'] is not None:
        cmd += " --layer_reference %(Cell2Location_layer_reference)s"
    if PARAMS['Cell2Location_categorical_covariate_keys_reference'] is not None:
        cmd += " --categorical_covariate_keys_reference %(Cell2Location_categorical_covariate_keys_reference)s"
    if PARAMS['Cell2Location_continuous_covariate_keys_reference'] is not None:
        cmd += " --continuous_covariate_keys_reference %(Cell2Location_continuous_covariate_keys_reference)s"
    if PARAMS['Cell2Location_max_epochs_reference'] is not None:
        cmd += " --max_epochs_reference %(Cell2Location_max_epochs_reference)s"
    
        # parameters for the spatial model
    if PARAMS['Cell2Location_batch_key_st'] is not None:
        cmd += " --batch_key_st %(Cell2Location_batch_key_st)s"
    if PARAMS['Cell2Location_layer_st'] is not None:
        cmd += " --layer_st %(Cell2Location_layer_st)s"
    if PARAMS['Cell2Location_categorical_covariate_keys_st'] is not None:
        cmd += " --categorical_covariate_keys_st %(Cell2Location_categorical_covariate_keys_st)s"
    if PARAMS['Cell2Location_continuous_covariate_keys_st'] is not None:
        cmd += " --continuous_covariate_keys_st %(Cell2Location_continuous_covariate_keys_st)s"
    if PARAMS['Cell2Location_max_epochs_st'] is not None:
        cmd += " --max_epochs_st %(Cell2Location_max_epochs_st)s"
    if PARAMS['Cell2Location_N_cells_per_location'] is not None:
        cmd += " --N_cells_per_location %(Cell2Location_N_cells_per_location)s"
    if PARAMS['Cell2Location_detection_alpha'] is not None:
        cmd += " --detection_alpha %(Cell2Location_detection_alpha)s"
    
    if PARAMS['Cell2Location_save_models'] is not None:
        cmd += " --save_models %(Cell2Location_save_models)s"
    
    cmd += " > logs/%(log_file)s "
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)




@follows(run_cell2location)
def full():
    """
    All cgat pipelines should end with a full() function which updates,
    if needed, all branches of the pipeline.
    The @follows statement should ensure that all functions are covered,
    either directly or as prerequisites.
    """
    pass


def main(argv=None):
    if argv is None:
        argv = sys.argv
    P.main(argv)


if __name__ == "__main__":
    sys.exit(P.main(sys.argv))



from ruffus import *
import sys
import os
from cgatcore import pipeline as P
import glob
import logging
from panpipes.funcs.io import dictionary_stripper



PARAMS = P.get_parameters(
    ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
     "pipeline.yml"])


PARAMS['py_path'] =  os.path.join(os.path.dirname(os.path.dirname(__file__)), 'python_scripts')
PARAMS['r_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'R_scripts')
PARAMS['resources_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")

    
job_kwargs = {}

if PARAMS['condaenv'] is not None:
    job_kwargs["job_condaenv"] =PARAMS['condaenv']
   



def gen_filter_jobs():
    if PARAMS["input_dir"] is not None:
        input_dir= PARAMS["input_dir"]
        if not os.path.exists(input_dir):
            logging.warning("input directory doesn't exist, checking if files are in the default ../qc.data/ dir")
            input_dir = "../qc.data"
            if not os.path.exists(input_dir):
                sys.exit("can't find input data")
    else:
        input_dir = "../qc.data"
        if not os.path.exists(input_dir):
                sys.exit("can't find input data")
    input_paths=glob.glob(os.path.join(input_dir,"*unfilt.h5mu"))
    for infile_path in input_paths:
        file_name = os.path.basename(infile_path)
        outfile = file_name.replace("unfilt","filtered")
        outfile = os.path.join("./filtered.data/",outfile)
        yield infile_path, outfile
    

@mkdir("logs")
@mkdir("figures")
@mkdir("tables")
@mkdir("filtered.data")
@files(gen_filter_jobs)
def filter_mudata(infile_path,outfile):
    print('processing file = %s' % str(infile_path))
    log_file = os.path.basename(outfile)
    log_file= "filtering."+log_file.replace("filtered.h5mu","") + ".log"


    filter_dict = dictionary_stripper(PARAMS['filtering'])
    cmd = """
        python %(py_path)s/run_filter_spatial.py
        --input_mudata %(infile_path)s
        --output_mudata %(outfile)s
        --filter_dict "%(filter_dict)s"
        """
    if PARAMS['filtering_keep_barcodes'] is not None:
        cmd += " --keep_barcodes %(filtering_keep_barcodes)s"
    cmd += " > logs/%(log_file)s "
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)
    


def run_plotqc_query(pqc_dict):
    # avoid deleting from PARAMS
    pqc = pqc_dict.copy()
    del pqc['grouping_var']
    return any([x != None for x in pqc.values()])


@active_if(run_plotqc_query(PARAMS['plotqc']))
@active_if(PARAMS['filtering_run'])
@transform(filter_mudata,
           regex("./filtered.data/(.*)_filtered.h5(.*)"), 
           r"./logs/postfilterplot.\1.log")
def postfilterplot_spatial(filt_file,log_file):
    print(filt_file)    
    print(log_file)
    spatial_filetype = PARAMS["assay"]
    cmd = """
            python %(py_path)s/plot_qc_spatial.py
             --input_mudata %(filt_file)s
             --spatial_filetype %(spatial_filetype)s
             --figdir ./figures/spatial
            """
    if PARAMS['plotqc']['grouping_var'] is not None:
        cmd += " --grouping_var %(plotqc_grouping_var)s"
    if PARAMS['plotqc']['spatial_metrics'] is not None:
        cmd += " --spatial_qc_metrics %(plotqc_spatial_metrics)s"
    cmd += " > %(log_file)s "
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)


@transform(filter_mudata,
           regex("./filtered.data/(.*)_filtered.h5(.*)"), 
           r"./logs/st_preprocess.\1.log")
def spatial_preprocess(filt_file,log_file):
    if os.path.exists("figures/spatial") is False:
        os.mkdir("figures/spatial")
    if os.path.exists("./tmp") is False:
        os.mkdir("./tmp")
    write_output = os.path.join("./tmp/",os.path.basename(filt_file))
    cmd = """
        python %(py_path)s/run_preprocess_spatial.py
        --input_mudata %(filt_file)s
        --output_mudata %(write_output)s
        --figdir ./figures/spatial
        """
    if PARAMS['spatial_norm_hvg_flavour'] is not None:
        cmd += " --norm_hvg_flavour %(spatial_norm_hvg_flavour)s"
    if PARAMS['spatial_n_top_genes'] is not None:
        cmd += " --n_top_genes %(spatial_n_top_genes)s"
    if PARAMS['spatial_filter_by_hvg'] is True:
        cmd += " --filter_by_hvg True"
    else:
        cmd += " --filter_by_hvg False"
    if PARAMS['spatial_hvg_batch_key'] is not None:
        cmd += " --hvg_batch_key %(spatial_hvg_batch_key)s"
    if PARAMS['spatial_squidpy_hvg_flavour'] is not None:
        cmd += " --squidpy_hvg_flavour %(spatial_squidpy_hvg_flavour)s"
    if PARAMS['spatial_min_mean'] is not None:
        cmd += " --min_mean %(spatial_min_mean)s"
    if PARAMS['spatial_max_mean'] is not None:
        cmd += " --max_mean %(spatial_max_mean)s"
    if PARAMS['spatial_min_disp'] is not None:
        cmd += " --min_disp %(spatial_min_disp)s"
    if PARAMS['spatial_theta'] is not None:
        cmd += " --theta %(spatial_theta)s"
    if PARAMS['spatial_clip'] is not None:
        cmd += " --clip %(spatial_clip)s"
    if PARAMS['spatial_n_pcs'] is not None:
        cmd += " --n_pcs %(spatial_n_pcs)s"

    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)

@follows(filter_mudata, postfilterplot_spatial, spatial_preprocess)
@originate("cleanup_done.txt")
def cleanup(file):
    # remove any ctmp fails
    P.run("rm ctmp*", without_cluster=True)
    #remove filtered folder to overwrite files
    P.run("rm -r filtered.data", without_cluster=True)
    #change tmp to filterd data
    P.run("mv tmp/ filtered.data", without_cluster=True)
    # remove tmp dir
    #P.run("rm -r tmp", without_cluster=True)
    # delete empty dirs
    P.run("find ./ -empty -type d -delete", without_cluster=True)


@follows(cleanup)
def full():
    """
    All cgat pipelines should end with a full() function which updates,
    if needed, all branches of the pipeline.
    The @follows statement should ensure that all functions are covered,
    either directly or as prerequisites.
    """
    pass

def main(argv=None):
    if argv is None:
        argv = sys.argv
    P.main(argv)


if __name__ == "__main__":
    sys.exit(P.main(sys.argv))



from anndata import AnnData
# from helpers import gen_adata
import panpipes.funcs as pnp
from scipy import sparse
import pytest
import pandas as pd
import numpy as np


@pytest.fixture()
def adata(nn=4):
    yield AnnData(sparse.eye(nn), 
        obs=pd.DataFrame(index=[f"cell{i}" for i in range(nn)]),
        var=pd.DataFrame(index=[f"gene{i}" for i in range(nn)]))


@pytest.fixture()
def mdata():
    yield MuData(
        {
            "mod1": AnnData(np.arange(0, 100, 0.1).reshape(-1, 10)),
            "mod2": AnnData(np.arange(101, 2101, 1).reshape(-1, 20)),
        }
    )

@pytest.fixture()
def anndata():
    yield AnnData(np.arange(50, 100, 1).reshape(-1, 10), 
        obs=pd.DataFrame(index=[f"cell{i}" for i in range(5)]),
        var=pd.DataFrame(index=[f"gene{i}" for i in range(10)]))

# def test_merge_with_adata_obs(adata):
#     adata['sample_id'] = 'a'
#     obs = pd.DataFrame(pd.DataFrame(index=[f"cell{i}" for i in range(4)]), data={'batch'==[0,0,1,1]})



'''
Plot spatial transcriptomics data
'''
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import scanpy as sc
import matplotlib.pyplot as plt
import seaborn as sns
import muon as mu
import os
import argparse
import sys
import logging
import re 
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("testing logger works")



sc.settings.verbosity = 3

parser = argparse.ArgumentParser()

parser.add_argument("--input_mudata",
                    default="mudata_unfilt.h5mu",
                    help="")
parser.add_argument("--figdir",
                    default="./figures/",
                    help="path to save the figures to")
parser.add_argument("--spatial_filetype",
                    default="",
                    help="")
parser.add_argument("--spatial_qc_metrics",
                    default="None",
                    help="metrics to plot")
parser.add_argument("--grouping_var",
                    default="None",
                    help="variables to group the cells by")



args, opt = parser.parse_known_args()

L.info("running with args:")
L.info(args)

figdir = args.figdir

if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))

mdata = mu.read(args.input_mudata)
spatial = mdata.mod['spatial']

input_data = os.path.basename(args.input_mudata)
pattern = r"_filtered.h5(.*)"
match = re.search(pattern, input_data)
if match is None:
    match = re.search(r"_unfilt.h5(.*)", input_data)
sprefix = input_data[:match.start()]

# convert string to list of strings
qc_metrics = list(args.spatial_qc_metrics.split(","))
group_var = list(args.grouping_var.split(","))


# check if metrics in adata.obs or adata.var
qc_metrics = [metric if metric in spatial.obs.columns or metric in spatial.var.columns else L.warning("Variable '%s' not found in adata.var or adata.obs, will not be plotted" % metric) for metric in qc_metrics]
qc_metrics = [metric for metric in qc_metrics if metric is not None]

# check that group_vars are in adata.obs
group_var = [group if group in spatial.obs.columns else L.warning("group_var '%s' not found in adata.obs, will be ignored" % group) for group in group_var]
group_var = [group for group in group_var if group is not None]
# make sure that it's saved as categorical 
for group in group_var: 
    spatial.obs[group] = spatial.obs[group].astype("category") 
    
if group_var == []:
    group_var = None


# plot the metrics

for metric in qc_metrics: 
    
    # check if in adata.obs:
    if metric in spatial.obs.columns: 
        # check that it's a numeric column, so that it can be plotted: 
        if metric not in spatial.obs._get_numeric_data().columns:
            L.warning("Variable '%s' not numerical in adata.obs, will not be plotted" % metric)
        else:
            if group_var is None: 
                sc.pl.violin(spatial, keys = metric, xlabel = metric+ " in .obs",
                            save =  "_obs_" + metric+ "_" + "."+sprefix + ".png", show = False)
            
            else: #plot violin for each group
                for group in group_var: 
                    sc.pl.violin(spatial, keys = metric,groupby = group, xlabel = group + ", "+ metric+ " in .obs",
                            save = "_obs_" + metric+ "_" + group+ "."+sprefix +".png", show = False)
            #plot spatial 
            sc.pl.embedding(spatial,basis="spatial", color = metric, save = "_spatial_" + metric + "."+sprefix +".png", show = False)

    #check if in adata.var: 
    if metric in spatial.var.columns:
        
        if metric not in spatial.var._get_numeric_data().columns:
            L.warning("Variable '%s' not numerical in adata.var, will not be plotted" % metric)
        else:
            # plot violins 
            ax = sns.violinplot(
                    data=spatial.var[[metric]],
                    orient='vertical', 
                )
            ax.set(xlabel=metric+ " in .var" )
            ax.figure.savefig(figdir + "/" +"violin_var_" + metric + "."+sprefix +".png")



if args.spatial_filetype == "vizgen":
    fig, axs = plt.subplots(1, 4, figsize=(15, 4))

    axs[0].set_title("Total transcripts per cell")
    sns.histplot(
        spatial.obs["total_counts"],
        kde=False,
        ax=axs[0],
    )

    axs[1].set_title("Unique transcripts per cell")
    sns.histplot(
        spatial.obs["n_genes_by_counts"],
        kde=False,
        ax=axs[1],
    )

    axs[2].set_title("Transcripts per FOV")
    sns.histplot(
        spatial.obs.groupby("fov").sum()["total_counts"],
        kde=False,
        ax=axs[2],
    )

    axs[3].set_title("Volume of segmented cells")
    sns.histplot(
        spatial.obs["volume"],
        kde=False,
        ax=axs[3],
    )

    plt.tight_layout()  # Ensures proper spacing between subplots
    #plt.savefig("merfish_histo.png", dpi=300)
    plt.savefig(figdir + "/histograms."+sprefix +".png", dpi=300)  # Adjust dpi as needed
    plt.close()  # Close the figure to free up memory

            

L.info("Done")




import argparse
import numpy as np
import pandas as pd
import muon as mu
import os
import re
import sys
import logging
import yaml

L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


parser = argparse.ArgumentParser()
parser.add_argument("--input_mudata")
parser.add_argument("--input_umap_files")
parser.add_argument('--rna_integration_col', default=None, help='the column in the adata.obs on which you have integrated')
parser.add_argument('--prot_integration_col', default=None, help='the column in the adata.obs on which you have integrated')
parser.add_argument('--atac_integration_col', default=None, help='the column in the adata.obs on which you have integrated')
parser.add_argument('--multimodal_integration_col', default=None, help='the column in the adata.obs on which you have integrated')
parser.add_argument("--output_cell_metadata_csv")
parser.add_argument("--output_combined_umaps_tsv")
parser.add_argument("--output_batch_yml")
args, opt = parser.parse_known_args()
L.info(args)
L.info("reading in files")

cell_meta_df = mu.read(args.input_mudata).obs
mtd_columns = cell_meta_df.columns.to_list()

# get all the batch columns 
# load up all the batch columns
batch_dict = dict(rna=args.rna_integration_col, 
                  prot=args.prot_integration_col, 
                  atac=args.atac_integration_col, 
                  multimodal=args.multimodal_integration_col)
batch_dict = {k:v.split(',') for k,v in batch_dict.items() if v is not None}

# format the column names correctly, and add any merge columns to the dataset
for k in batch_dict.keys():
    v=batch_dict[k]
    print(k)
    print(v)
    if k =="multimodal":
        mk = []
        exc = []
        if isinstance(v, str):
            if v in cell_meta_df.columns:
                mk.append(v)
            else:
                exc.append(v)
        else:
            for v2 in (v):
                if v2 in cell_meta_df.columns:
                    mk.append(v2)
                else:
                    exc.append(v2)    
        if len(exc)>0:
            print(exc)
            tmp =[]
            for g in ["atac","rna","prot","rep"]:
                for e in exc:
                    md = f"{g}:{e}"  # Construct the modality:dataset string
                    tmp.append(md)
            for t in tmp:
                if t in cell_meta_df.columns:
                    mk.append(t)
        v = mk
        batch_dict[k] = v
    else:
        v =[k + ":" +v1 if k!="multimodal" else v1 for v1 in v  ] #add the prepending modality
        batch_dict[k] = v
    if len(v) > 1:
        print(v)
        mod_meta_df= cell_meta_df[v].dropna()
        # Create a list to store identical column pairs
        identical_column_pairs = []

        # Iterate through pairs of columns and check if they are identical
        for i in range(len(v)):
            for j in range(i + 1, len(v)):
                col1 = v[i]
                col2 = v[j]
                
                if mod_meta_df[col1].equals(mod_meta_df[col2]):
                    identical_column_pairs.append((col1, col2))
        if identical_column_pairs:
            print("Identical column pairs:")
            for col1, col2 in identical_column_pairs:
                print(f"{col1} and {col2} are identical.")
            batch_dict[k].append(col1)
        else:
            print("No identical columns found.")
            cell_meta_df[str(k)+ ":bc_batch"] = mod_meta_df.apply(lambda x: '|'.join(x), axis=1)
            batch_dict[k].append(k+ ":bc_batch")
    L.info("batch values %s" %(batch_dict[k]))

for k in batch_dict.keys():
    unique_values = list(set(batch_dict[k]))
    batch_dict[k] = unique_values

L.info("saving batch dictionary")
print(batch_dict)

cell_meta_df.to_csv(args.output_cell_metadata_csv)

with open(args.output_batch_yml, 'w') as outfile:
    yaml.dump(batch_dict, outfile, default_flow_style=False)

L.info("reading in all umaps")
# load files
input_files = args.input_umap_files.split(",")
umaps_list = [pd.read_csv(x, index_col=0) for x in input_files]
umaps_index = [re.sub("umap_|.csv", "", os.path.basename(x)) for x in input_files]
umaps_df = pd.concat(umaps_list, axis=0, keys=umaps_index).reset_index(level=0)
umaps_df.columns = ["method", "umap_1", "umap_2"]
umaps_df = umaps_df.merge(cell_meta_df, left_index=True, right_index=True)

# reorder methods so None is at the most left on plot
umaps_df['mod'] = [x.split('_')[0] for x in umaps_df.method]
umaps_df['method'] = [x.split('_')[1] for x in umaps_df.method]

# tidy up umaps df before writing out
def move_column_inplace(df, col, pos):
    col = df.pop(col)
    df.insert(pos, col.name, col)
move_column_inplace(umaps_df, "mod", 0)
move_column_inplace(umaps_df, "method", 1)
umaps_df = umaps_df.sort_values(by=['mod','method'])

# save umap df to file
L.info("save umap df to file")
umaps_df = umaps_df.iloc[:,0:4]
umaps_df['cellbarcode'] = umaps_df.index
umaps_df.to_csv(args.output_combined_umaps_tsv, sep="\t")

L.info("done")




'''
Run Cell2Location
'''
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import cell2location as c2l
import scanpy as sc
import pandas as pd
import muon as mu

import os
import argparse
import sys
import logging

from panpipes.funcs.plotting import cell2loc_plot_QC_reference
from panpipes.funcs.plotting import cell2loc_plot_QC_reconstr
from panpipes.funcs.plotting import cell2loc_plot_history
from panpipes.funcs.scmethods import cell2loc_filter_genes


L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("testing logger works")



sc.settings.verbosity = 3

parser = argparse.ArgumentParser()

parser.add_argument("--input_spatial",
                    help="path to mudata of spatial transcriptomics data")
parser.add_argument("--input_singlecell",
                    help="path to mudata of single-cell reference data")
parser.add_argument("--figdir",
                    default="./figures/Cell2Location",
                    help="path to save the figures to")
parser.add_argument("--output_dir",
                    default="./cell2location.output",
                    help="path to save the figures to")
parser.add_argument("--save_models",
                    default=False,
                    help="whether to save the reference & spatial mapping models")


# parameters for feature selection: 
parser.add_argument("--gene_list",
                    default="None",
                    help="path to a csv containing a list of genes to use for the deconvolution")
parser.add_argument("--remove_mt",
                    default=True,
                    help="whether to remove MT genes before deconvolution")
parser.add_argument("--cell_count_cutoff",
                    default=15,
                    help="all genes detected in less than cell_count_cutoff cells will be excluded")
parser.add_argument("--cell_percentage_cutoff2",
                    default=0.05,
                    help="all genes detected in at least this percentage of cells will be included")
parser.add_argument("--nonz_mean_cutoff",
                    default=1.12,
                    help="genes detected in the number of cells between the above mentioned cutoffs are selected only when their average expression in non-zero cells is above this cutoff")

# parameters for reference model: 
parser.add_argument("--labels_key_reference",
                    default=None,
                    help="")
parser.add_argument("--batch_key_reference",
                    default=None,
                    help="")
parser.add_argument("--layer_reference",
                    default=None,#or raw_counts
                    help="")
parser.add_argument("--categorical_covariate_keys_reference",
                    default=None,
                    help="")
parser.add_argument("--continuous_covariate_keys_reference",
                    default=None,
                    help="")
parser.add_argument("--max_epochs_reference",
                    default=None,
                    help="")

# parameters for spatial model: 
parser.add_argument("--labels_key_st",
                    default=None,
                    help="")
parser.add_argument("--batch_key_st",
                    default=None,
                    help="")
parser.add_argument("--layer_st",
                    default=None,#or raw_counts
                    help="")
parser.add_argument("--categorical_covariate_keys_st",
                    default=None,
                    help="")
parser.add_argument("--continuous_covariate_keys_st",
                    default=None,
                    help="")
parser.add_argument("--N_cells_per_location",
                    default=None,
                    help="")
parser.add_argument("--detection_alpha",
                    default=None,
                    help="")
parser.add_argument("--max_epochs_st",
                    default=None,
                    help="")  


args, opt = parser.parse_known_args()

L.info("running with args:")
L.debug(args)

figdir = args.figdir
if not os.path.exists(figdir):
    os.mkdir(figdir)
sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))

output_dir = args.output_dir
if not os.path.exists(output_dir):
    os.mkdir(output_dir)


if args.N_cells_per_location is None: 
    L.error("The parameter 'N_cells_per_location' is not specified. The parameters 'N_cells_per_location' and 'detection_alpha' need to be specified!")
    sys.exit("The parameter 'N_cells_per_location' is not specified. The parameters 'N_cells_per_location' and 'detection_alpha' need to be specified!")
if args.detection_alpha is None: 
    L.error("The parameter 'detection_alpha' is not specified. The parameters 'N_cells_per_location' and 'detection_alpha' need to be specified!")
    sys.exit("The parameter 'detection_alpha' is not specified. The parameters 'N_cells_per_location' and 'detection_alpha' need to be specified!")
    
    
if (args.save_models is False) or (args.save_models == "False"): 
    save_models = False
else:
    save_models = True
      
if (args.remove_mt is True) or (args.remove_mt == "True"): 
    remove_mt = True
else:
    remove_mt = False


if args.categorical_covariate_keys_reference is not None: 
    categorical_covariate_keys_reference = list(args.categorical_covariate_keys_reference.split(","))
else:
    categorical_covariate_keys_reference = None
    
if args.categorical_covariate_keys_st is not None: 
    categorical_covariate_keys_st = list(args.categorical_covariate_keys_st.split(","))
else:
    categorical_covariate_keys_st = None
    
if args.continuous_covariate_keys_reference is not None: 
    continuous_covariate_keys_reference = list(args.continuous_covariate_keys_reference.split(","))
else:
    continuous_covariate_keys_reference = None
    
if args.continuous_covariate_keys_st is not None: 
    continuous_covariate_keys_st = list(args.continuous_covariate_keys_st.split(","))
else:
    continuous_covariate_keys_st = None

    
if args.max_epochs_reference is None: 
    max_epochs_reference = None
else: 
    max_epochs_reference = int(args.max_epochs_reference)
    
if args.max_epochs_st is None: 
    max_epochs_st = None
else: 
    max_epochs_st= int(args.max_epochs_st)




#1. read in the data
#spatial: 
mdata_spatial = mu.read(args.input_spatial)
adata_st = mdata_spatial.mod['spatial']
#single-cell: 
mdata_singlecell = mu.read(args.input_singlecell)
adata_sc = mdata_singlecell.mod['rna']



#2. Perform gene selection:
if args.gene_list != "None": # read in csv and subset both anndatas
    reduced_gene_set = pd.read_csv(args.gene_list, header = 0)
    reduced_gene_set.columns = ["HVGs"]
    adata_sc.var["selected_gene"] = adata_sc.var.index.isin(reduced_gene_set["HVGs"])
    adata_st.var["selected_gene"] = adata_st.var.index.isin(reduced_gene_set["HVGs"])
    adata_sc = adata_sc[:, adata_sc.var["selected_gene"]]
    adata_st = adata_st[:, adata_st.var["selected_gene"]]
    # check whether all genes are present in both, spatial & reference
    if set(adata_st.var.index) != set(adata_sc.var.index):
        L.error(
            "Not all genes of the gene list %s are present in the reference as well as in the ST data. Please provide a gene list where all genes are present in both, reference and ST.", args.gene_list)
        sys.exit(
            "Not all genes of the gene list are present in the reference as well as in the ST data. Please provide a gene list where all genes are present in both, reference and ST.")

else: # perform feature selection according to cell2loc
    if remove_mt is True: 
        adata_st.var["MT_gene"] = [gene.startswith("MT-") for gene in adata_st.var.index]
        adata_st.obsm["MT"] = adata_st[:, adata_st.var["MT_gene"].values].X.toarray()
        adata_st = adata_st[:, ~adata_st.var["MT_gene"].values]
    # intersect vars of reference and spatial
    shared_features = [feature for feature in adata_st.var_names if feature in adata_sc.var_names]
    adata_sc = adata_sc[:, shared_features]
    adata_st = adata_st[:, shared_features]
    # select features
    selected = cell2loc_filter_genes(adata_sc, figdir + "/gene_filter.png", cell_count_cutoff=float(args.cell_count_cutoff),
                                               cell_percentage_cutoff2=float(args.cell_percentage_cutoff2),
                                                nonz_mean_cutoff=float(args.nonz_mean_cutoff))

    adata_sc = adata_sc[:, selected]
    adata_st = adata_st[:, selected]

    

# 3. Fit regression model 
c2l.models.RegressionModel.setup_anndata(adata=adata_sc, 
                                         labels_key = args.labels_key_reference,
                                         layer= args.layer_reference, 
                                         batch_key= args.batch_key_reference,
                                         categorical_covariate_keys = categorical_covariate_keys_reference,
                                         continuous_covariate_keys =  continuous_covariate_keys_reference)
model_ref = c2l.models.RegressionModel(adata_sc)
model_ref.train(max_epochs=max_epochs_reference)

# plot elbo
cell2loc_plot_history(model_ref, figdir + "/ELBO_reference_model.png")

# export results
adata_sc = model_ref.export_posterior(adata_sc)
if "means_per_cluster_mu_fg" in adata_sc.varm.keys():
    inf_aver = adata_sc.varm["means_per_cluster_mu_fg"][[f"means_per_cluster_mu_fg_{i}" for i in adata_sc.uns["mod"]["factor_names"]]].copy()
else:
    inf_aver = adata_sc.var[[f"means_per_cluster_mu_fg_{i}" for i in adata_sc.uns["mod"]["factor_names"]]].copy()
inf_aver.columns = adata_sc.uns["mod"]["factor_names"]
inf_aver.to_csv(output_dir+"/Cell2Loc_inf_aver.csv")

# plot QC
cell2loc_plot_QC_reference(model_ref, figdir + "/QC_reference_reconstruction_accuracy.png", figdir + "/QC_reference_expression signatures_vs_avg_expression.png")

# save model and update mudata
mdata_singlecell.mod["rna"] = adata_sc
mdata_singlecell.update()
if save_models is True:
    model_ref.save(output_dir +"/Reference_model", overwrite=True)


       
# 4. Fit mapping model   
c2l.models.Cell2location.setup_anndata(adata=adata_st, 
                                         labels_key = args.labels_key_st,
                                         layer= args.layer_st, 
                                         batch_key= args.batch_key_st,
                                         categorical_covariate_keys = categorical_covariate_keys_st,
                                         continuous_covariate_keys =  continuous_covariate_keys_st)

        
model_spatial = c2l.models.Cell2location(adata = adata_st, cell_state_df=inf_aver, 
                                        N_cells_per_location=float(args.N_cells_per_location),
                                        detection_alpha=float(args.detection_alpha))
model_spatial.train(max_epochs=max_epochs_st)

#plot elbo
cell2loc_plot_history(model_spatial, figdir + "/ELBO_spatial_model.png")
#extract posterior
adata_st = model_spatial.export_posterior(adata_st)
#plot QC
cell2loc_plot_QC_reconstr(model_spatial, figdir + "/QC_spatial_reconstruction_accuracy.png")


#plot output
adata_st.obs[adata_st.uns["mod"]["factor_names"]] = adata_st.obsm["q05_cell_abundance_w_sf"]
sc.pl.spatial(adata_st,color=adata_st.uns["mod"]["factor_names"], show = False, save = "_Cell2Loc_q05_cell_abundance_w_sf.png") 


# save model and update mudata
mdata_spatial.mod["spatial"] = adata_st
mdata_spatial.update()
if save_models is True: 
    model_spatial.save(output_dir+"/Spatial_mapping_model", overwrite=True)


#6. save mudatas 
mdata_singlecell.write(output_dir+"/Cell2Loc_screference_output.h5mu")
mdata_spatial.write(output_dir+"/Cell2Loc_spatial_output.h5mu")


L.info("Done")




import multiprocessing 

import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import scvi
import argparse
import os
import gc
import muon as mu
from cgatcore import pipeline as P
import anndata as ad
import panpipes.funcs as pp
import gc
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice, X_is_raw

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--scaled_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_MultiVI.csv',
                    help='')
parser.add_argument('--integration_col_categorical', default='batch',
                    help='')
parser.add_argument('--integration_col_continuous', default=None,
                    help='')
parser.add_argument('--figdir', default='./figures',
                    help='')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs", default=50)
parser.add_argument('--neighbors_method', default="scanpy",
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k', default=30,
                    help="neighbors k")
parser.add_argument('--neighbors_metric',default="euclidean",
                    help="neighbor metric, e.g. euclidean or cosine")




args, opt = parser.parse_known_args()

# scanpy settings
sc.set_figure_params(facecolor="white")
sc.settings.autoshow = False
sc.settings.figdir = args.figdir


# load parameters

threads_available = multiprocessing.cpu_count()
params = pp.io.read_yaml("pipeline.yml")

test_script=False
if test_script:
    L.info("this is a test run")
    params['MultiVI']['training_args']['max_epochs'] = 10

# ------------------------------------------------------------------
L.info("Running multivi")

# mdata = mu.read(args.scaled_anndata)
# rna = mdata['rna'].copy()
# atac = mdata['atac'].copy()
rna= mu.read(args.scaled_anndata +"/" + "rna")
atac= mu.read(args.scaled_anndata +"/" + "atac")


if check_for_bool(params["multimodal"]["MultiVI"]["lowmem"]):
    L.info("subsetting atac to top 25k HVF")
    sc.pp.highly_variable_genes(atac, n_top_genes=25000)
    atac = atac[:, atac.var.highly_variable]



gc.collect()

# if "modality" not in atac.var.keys():
#     atac.var["modality"] = "Peaks"

# if "modality" not in rna.var.keys():
#     rna.var["modality"] = "Gene Expression"

if rna.shape[0] == atac.shape[0]:
    n=int(rna.shape[0])
else:
    sys.exit("rna and atac have different number of cells, \
        can't deal with this in this version of MultiVI integration")

gc.collect()

n_genes = len(rna.var_names)
n_regions = len(atac.var_names)


if "raw_counts" in rna.layers.keys():
    L.info("raw counts found in layer for RNA")
elif X_is_raw(rna):
    # this means the X layer is already raw and we can make the layer we need
    L.info("raw counts found in X for RNA")
    rna.layers["raw_counts"] = rna.X.copy()
else:
    sys.exit("please provide a mu/anndata with raw counts for RNA")

if "raw_counts" in atac.layers.keys():
    L.info("raw counts found in layer for ATAC")
elif X_is_raw(atac):
    # this means the X layer is already raw and we can make the layer we need
    L.info("raw counts found in X for ATAC")
    atac.layers["raw_counts"] = atac.X.copy()
else:
    sys.exit("please provide a mu/anndata with raw counts for ATAC")

L.info("concatenating modalities to comply with multiVI")
# adata_paired = ad.concat([rna, atac], join="outer")
# adata_paired.var = pd.concat([rna.var,atac.var])

adata_paired = ad.concat([rna.T, atac.T]).T

rna_cols=rna.obs.columns
atac_cols=atac.obs.columns

rnaobs = rna.obs.copy()
rnaobs.columns= ["rna:"+ x for x in rna_cols]
atacobs = atac.obs.copy()
atacobs.columns= ["atac:"+ x for x in atac_cols]
adata_paired.obs = pd.merge(rnaobs, atacobs, left_index=True, right_index=True)

if "modality" not in adata_paired.obs.columns:
    adata_paired.obs["modality"] = "paired" 

print(adata_paired)

# adata = ad.concat([rna,atac],join="outer")
# adata.var = pd.concat([rna.var,atac.var])

del [rna , atac ]
gc.collect()

adata_mvi = scvi.data.organize_multiome_anndatas(adata_paired)


# MultiVI integrates by modality, to use batch correction you need a batch covariate to specify in
# categorical_covariate_keys
if args.integration_col_categorical is not None :
    cols = [x.strip() for x in args.integration_col_categorical.split(",")]
    columns = []
    for cc in cols:
        if cc in rna_cols:
            print("categorical batch covariate in rna")
            columns.append("rna:"+cc)
        elif cc in atac_cols:
            print("categorical batch covariate in atac")
            columns.append("atac:"+cc)
if args.integration_col_continuous is not None :
    if args.integration_col_continuous in rna_cols:
        args.integration_col_continuous = "rna:"+ args.integration_col_continuous
    elif args.integration_col_continuous in atac_cols:
        args.integration_col_continuous = "atac:"+ args.integration_col_continuous


kwargs = {}
# in case of more than 1 variable, create a fake column with combined information
if columns is not None:
    print(columns)
    if len(columns) > 1:
        L.info("using 2 columns to integrate on more variables")
        # bc_batch = "_".join(columns)
        adata_mvi.obs["bc_batch"] = adata_mvi.obs[columns].apply(lambda x: '|'.join(x), axis=1) 
        # make sure that batch is a categorical
        adata_mvi.obs["bc_batch"] = adata_mvi.obs["bc_batch"].astype("category")
    else:
        adata_mvi.obs['bc_batch'] = adata_mvi.obs[columns]
        adata_mvi.obs["bc_batch"] = adata_mvi.obs["bc_batch"].astype("category")

    batch_categories = list(adata_mvi.obs['bc_batch'].unique())
    kwargs['categorical_covariate_keys'] = ["bc_batch"]

if args.integration_col_continuous is not None :
    print(args.integration_col_continuous)
    adata_mvi.obs['bc_batch_continuous'] = adata_mvi.obs[args.integration_col_continuous]
    kwargs['continuous_covariate_keys'] = ["bc_batch_continuous"]

print(kwargs)

# 1 setup anndata
#scvi.model.MULTIVI.setup_anndata(adata_mvi, batch_key='modality', **kwargs)
scvi.model.MULTIVI.setup_anndata(
    adata_mvi, 
    batch_key='modality',
    layer =  "raw_counts",
    **kwargs) 
#2. setup model
if params["multimodal"]['MultiVI']['model_args'] is None:
    multivi_model_args =  {}
else:
    multivi_model_args =  {k: v for k, v in params["multimodal"]['MultiVI']['model_args'].items() if v is not None}

print(multivi_model_args)

mvi = scvi.model.MULTIVI(
    adata_mvi,
    n_genes=n_genes,
    n_regions=n_regions,
    **multivi_model_args
)

#3.train

if params["multimodal"]["MultiVI"]["training_args"] is None:
    multivi_training_args={}
else:
    multivi_training_args =  {k: v for k, v in params["multimodal"]['MultiVI']['training_args'].items() if v is not None}

print(multivi_training_args)

if params["multimodal"]['MultiVI']['training_plan'] is None:
    multivi_training_plan = {}
else:
    multivi_training_plan =  {k: v for k, v in params["multimodal"]['MultiVI']['training_plan'].items() if v is not None}

print(multivi_training_plan)

mvi.view_anndata_setup()

mvi.train( **multivi_training_args, **multivi_training_plan)

mvi.save(os.path.join("batch_correction", "multivi_model"), 
                  anndata=False)

plt.plot(mvi.history["elbo_train"], label="train")
plt.plot(mvi.history["elbo_validation"], label="validation")
plt.title("Negative ELBO over training epochs")

plt.legend()
plt.savefig(os.path.join(args.figdir, "multivi_elbo_plot.png"))

L.info("""We support the use of mudata as a general framework for multimodal data
        For this reason, the object we save is not the classical anndata
        you would find in scvitools tutorial.
        We chose to save the learned SC representation in the 
        Mudata obsm slot, and any single modality processing in
        its own modality slot
            """)

mdata = mu.read(args.scaled_anndata)
mdata.obsm["X_MultiVI"] = mvi.get_latent_representation()
#adata_mvi.obsm["X_MultiVI"] = mvi.get_latent_representation()

if int(args.neighbors_n_pcs) > mdata.obsm['X_MultiVI'].shape[1]:
    L.warn(f"N PCs is larger than X_MultiVI dimensions, reducing n PCs to  {mdata.obsm['X_MultiVI'].shape[1] -1}")
n_pcs= min(int(args.neighbors_n_pcs), mdata.obsm['X_MultiVI'].shape[1]-1)

run_neighbors_method_choice(mdata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=n_pcs, 
    metric=args.neighbors_metric, 
    use_rep='X_MultiVI',
    nthreads=max([threads_available, 6]))
sc.tl.umap(mdata, min_dist=0.4)
sc.tl.leiden(mdata, key_added="leiden_multiVI")

umap = pd.DataFrame(mdata.obsm['X_umap'], mdata.obs.index)
umap.to_csv(args.output_csv)

write_anndata(mdata, "tmp/multivi_scaled_adata.h5mu",use_muon=False)

L.info("Done")




## Filtering mudata based on n_genes, percent mito and min cells/
## originally written by Tom Thomas (https://github.com/tomthomas3000/TAURUS)
## adapted and augmented for this pipeline by Charlotte Rich-Griffin 2020-09-30

import scanpy as sc
import argparse
import pandas as pd
import re
import muon as mu
from anndata import AnnData
from muon import MuData
import yaml

# import scpipelines.funcs as scp
from panpipes.funcs.processing import intersect_obs_by_mod, remove_unused_categories
from panpipes.funcs.io import write_obs, read_yaml, dictionary_stripper
import collections.abc
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


sc.settings.verbosity = 2  # verbosity: errors (0), warnings (1), info (2), hints (3)
# comment this because of numba issues
# sc.logging.L.info_versions()



def map_nested_dicts_remove_none(ob):
    if isinstance(ob, collections.abc.Mapping):
        return {k: map_nested_dicts_remove_none(v) for k, v in ob.items() if v is not None}
    else:
        return ob


def test_matching_df_ignore_cat(new_df, old_df):
    remove_unused_categories(old_df)
    old_df = old_df.loc[new_df.index, :]
    remove_unused_categories(old_df)
    return new_df.equals(new_df)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_mudata',
                    default='gut_minus1_amp.h5ad',
                    help='')
parser.add_argument('--output_mudata',
                    default='',
                    help='')
parser.add_argument('--filter_dict',
                    default='',
                    help='this is pull')
# cross modalities args
parser.add_argument('--keep_barcodes', default=None,
                    help='1 column list of barcodes to keep, note that they should match the mudata input, this filtering happens first')
parser.add_argument('--intersect_mods', default=None,
                    help='comma separated string of modalities we want to intersect_obs on')                 


# load options
parser.set_defaults(verbose=True)
args, opt = parser.parse_known_args()
L.info('running with args:')
L.info(args)
 
# load the filtering dictionary 
filter_dict = args.filter_dict
if isinstance(args.filter_dict, dict):
    filter_dict = args.filter_dict
else:
    filter_dict = read_yaml(args.filter_dict) 

# remove Nones for each level of filter dict
filter_dict = map_nested_dicts_remove_none(filter_dict)
filter_dict = dictionary_stripper(filter_dict)
L.info("filter dictionary:\n %s" %filter_dict)

# load mudata
mdata = mu.read(args.input_mudata)

if isinstance(mdata, AnnData):
    raise TypeError("input_mudata should be of MuData format, not Anndata")

orig_obs = mdata.obs.copy()

L.info("Before filtering %d" % mdata.n_obs)
# filter based on provided barcodes -----
if args.keep_barcodes is not None:
    L.info("filtering all modalities by keep_barcodes file")
    keep_bc = pd.read_csv(args.keep_barcodes,header=None)
    mdata = mdata[mdata.obs_names.isin(keep_bc[0]),:].copy()
    remove_unused_categories(mdata.obs)
    mdata.update()
    L.info("Remaining cells %d" % mdata.n_obs)

# L.debug(mdata.obs['sample_id'].value_counts())

# filter more than
if filter_dict['run']:
    # this will go through the modalities one at a time,
    # then the categories max, min and bool
    for mod in mdata.mod.keys():
        L.info(mod)
        if mod in filter_dict.keys():
            for marg in filter_dict[mod].keys():
                if marg == "obs":
                    if "max" in filter_dict[mod][marg].keys():
                        for col, n in filter_dict[mod][marg]['max'].items():
                            L.info("filter %s %s to less than %s" % (mod, col, n))
                            mu.pp.filter_obs(mdata.mod[mod], col, lambda x: x <= n)
                            L.info("Remaining cells %d" % mdata[mod].n_obs)
                            L.info("Remaining cells %d" % mdata[mod].shape[0])
                    if "min" in filter_dict[mod][marg].keys():
                        for col, n in filter_dict[mod][marg]['min'].items():
                            L.info("filter %s %s to more than %s" % (mod, col, n))
                            mu.pp.filter_obs(mdata.mod[mod], col, lambda x: x >= n)
                            L.info("Remaining cells %d" % mdata[mod].n_obs)
                    if "bool" in filter_dict[mod][marg].keys():
                        for col, n in filter_dict[mod][marg]['bool'].items():
                            L.info("filter %s %s marked %s" % (mod, col, n))
                            mu.pp.filter_obs(mdata.mod[mod], col, lambda x: x == n)
                            L.info("Remaining cells %d" % mdata[mod].n_obs)
                if marg == "var":
                    if "max" in filter_dict[mod][marg].keys():
                        for col, n in filter_dict[mod][marg]['max'].items():
                            L.info("filter %s %s to less than %s" % (mod, col, n))
                            mu.pp.filter_var(mdata.mod[mod], col, lambda x: x <= n)
                            L.info("Remaining cells %d" % mdata[mod].n_obs)

                    if "min" in filter_dict[mod][marg].keys():
                        for col, n in filter_dict[mod][marg]['min'].items():
                            L.info("filter %s %s to more than %s" % (mod, col, n))
                            mu.pp.filter_var(mdata.mod[mod], col, lambda x: x >= n)
                            L.info("Remaining cells %d" % mdata[mod].n_obs)

                    if "bool" in filter_dict[mod][marg].keys():
                        for col, n in filter_dict[mod][marg]['bool'].items():
                            L.info("filter %s %s marked %s" % (mod, col, n))
                            mu.pp.filter_var(mdata.mod[mod], col, lambda x: x == n)
                            L.info("Remaining cells %d" % mdata[mod].n_obs)
                            
L.debug(mdata.obs['sample_id'].value_counts())
L.debug(mdata['rna'].obs['sample_id'].value_counts())
mdata.update()
L.debug(mdata.obs['sample_id'].value_counts())
L.debug(mdata['rna'].obs['sample_id'].value_counts())
# intersect specific modalities
# we don't want to just use mu.pp.intersect_obs, 
# because we don't want to necessarily filter by repertoire
if args.intersect_mods is not None:
    intersect_mods = args.intersect_mods.split(',')
    intersect_mods= [a.strip() for a in intersect_mods]
    L.info("intersecting barcodes in %s" % args.intersect_mods)
    intersect_obs_by_mod(mdata, mods = intersect_mods)
    L.info("Remaining cells %d" % mdata.n_obs)
    L.info(mdata.shape)



remove_unused_categories(mdata.obs)

# run quick test before saving out.
assert test_matching_df_ignore_cat(mdata.obs, orig_obs)  

# write out obs
output_prefix = re.sub(".h5mu", "", args.output_mudata)
write_obs(mdata, output_prefix=output_prefix, output_suffix="_filtered_cell_metadata.tsv")

# write out the per sample_id cell numbers 
cell_counts_dict={}
for mm in mdata.mod.keys():
    cell_counts_dict[mm] = mdata[mod].obs.sample_id.value_counts().to_frame('n_cells')

cell_counts = pd.concat(cell_counts_dict).reset_index().rename(
    columns={"level_0": "modality", "level_1": "sample_id"})

L.info("cell_counts\n%s" %cell_counts)
cell_counts.to_csv(output_prefix + "_cell_counts.csv", index=None)


mdata.update()
L.info("writing mdata h5mu")
L.info(args.output_mudata)
L.debug(mdata.obs['sample_id'].value_counts())
L.debug(mdata['rna'].obs['sample_id'].value_counts())
mdata.write(args.output_mudata)

L.info("Done")




## Processing mudata based on n_genes, percent mito and min cells/
## originally written by Tom Thomas (https://github.com/tomthomas3000/TAURUS)
## adapted for this pipeline by Charlotte Rich-Griffin 2020-09-30


# import numpy as np
import pandas as pd
import numpy as np
import scanpy as sc
# import scipy.io
# import matplotlib
# matplotlib.use("Agg")
# import matplotlib.pyplot as plt
# plt.ioff()
import os
import argparse
import re
import muon as mu


from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.scmethods import X_is_raw

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()


parser.add_argument('--input_mudata',
                    default='data/mudata-n5000_filt.h5ad',
                    help='')
parser.add_argument('--output_scaled_mudata',
                    default=None,
                    help='if not specified then the output will be written to output_logged_mudata path')
parser.add_argument('--use_muon',
                    default=False, type=check_for_bool,
                    help='')  
parser.add_argument('--fig_dir', 
                    default="./figures",
                    help="save plots here")
parser.add_argument('--exclude_file', default=None,
                    help='')
parser.add_argument('--exclude', default="exclude",
                    help='how are genes that need to be excl. from HVG are labeled')

# highly variable genes options
parser.add_argument('--flavor', default='seurat')
parser.add_argument('--n_top_genes', default=None)
parser.add_argument('--min_mean', default=0.0125)
parser.add_argument('--max_mean', default=3)
parser.add_argument('--min_disp', default=0.5)
parser.add_argument("--filter_by_hvg", default=False, type=check_for_bool)
parser.add_argument('--hvg_batch_key', default=None)
# regress out options
parser.add_argument('--regress_out', default=None)
# scale options
parser.add_argument('--scale', default=True, type=check_for_bool)
parser.add_argument('--scale_max_value', default=None)
# pca options
parser.add_argument("--n_pcs", default=50)
parser.add_argument("--pca_solver", default="arpack")
parser.add_argument("--color_by", default="batch") 

parser.set_defaults(verbose=True)
args, opt = parser.parse_known_args()
L.info(args)

figdir = args.fig_dir
if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))

mdata = mu.read(args.input_mudata)

# if we have actually loaded an anndata object, make into a mudata
if isinstance(mdata, sc.AnnData):
    mdata =  mu.MuData({'rna': mdata})

adata = mdata['rna']
# resolve multi-column batch for hvg batch key
if args.hvg_batch_key is not None:
    columns = [x.strip() for x in args.hvg_batch_key.split(",")]
    if len(columns) > 1: 
        L.info("combining batch columns into one column 'hvg_batch_key'")
        adata.obs["hvg_batch_key"] = adata.obs[columns].apply(lambda x: '|'.join(x), axis=1)
        # make sure that batch is a categorical
        adata.obs["hvg_batch_key"] = adata.obs["hvg_batch_key"].astype("category")
        hvg_batch_key="hvg_batch_key"
    else:
        hvg_batch_key=columns[0]
else:
    hvg_batch_key=None

# save raw counts as a layer
if X_is_raw(adata):
    adata.layers['raw_counts'] = adata.X.copy()
elif "raw_counts" in adata.layers :
    L.info("raw_counts layer already exists")
    adata.X = adata.layers['raw_counts'].copy()
else:
    L.error("X is not raw data and raw_counts layer not found")
    sys.exit("X is not raw data and raw_counts layer not found")


# sc.pp.highly variable genes Expects logarithmized data, 
# except when flavor='seurat_v3' in which count data is expected.
# change the order accordingly
if X_is_raw(adata):
    L.info("normalise, log and calculate highly variable genes")
    if args.flavor == "seurat_v3":
        if args.n_top_genes is None:
            raise ValueError("if seurat_v3 is used you must give a n_top_genes value")
        else:
            sc.pp.highly_variable_genes(adata, flavor="seurat_v3", 
                                        n_top_genes=int(args.n_top_genes),
                                        batch_key=hvg_batch_key)
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
    else:
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        L.debug(adata.uns['log1p'])
        sc.pp.highly_variable_genes(adata, flavor=args.flavor,
                                    min_mean=float(args.min_mean), 
                                    max_mean=float(args.max_mean),
                                    min_disp=float(args.min_disp),
                                    batch_key=hvg_batch_key)
        L.debug(adata.uns['log1p'])
if "highly_variable" in adata.var: 
    L.warning( "You have %s Highly Variable Features", np.sum(adata.var.highly_variable))
    sc.pl.highly_variable_genes(adata,show=False, save ="_genes_highlyvar.png")
else:
    sys.exit("I cannot find Highly Variable Features in your RNA")

if isinstance(mdata, mu.MuData):
    mdata.update()

# exclude hvgs if there is an exclude file
if args.exclude_file is not None:
    if os.path.exists(args.exclude_file):
        L.info("exclude genes from hvg")
        customgenes = pd.read_csv(args.exclude_file) 
        custom_cat = list(set(customgenes['group'].tolist()))
        cat_dic = {}
        for cc in custom_cat:
            cat_dic[cc] = customgenes.loc[customgenes["group"] == cc,"feature"].tolist()
        exclude_action = str(args.exclude)
        excl = cat_dic[exclude_action]
        L.info(len(excl))
        L.info("number of hvgs prior to filtering")
        L.info(adata.var.highly_variable.sum())
        adata.var['hvg_exclude'] = [True if gg in excl else False for gg in adata.var.index]
        L.debug("num to exclude %s" %  (adata.var['hvg_exclude'] & adata.var['highly_variable']).sum() )
        L.debug(adata)
        if any(adata.var['hvg_exclude'] & adata.var['highly_variable']):
            adata.var['highly_variable'].loc[adata.var.hvg_exclude] = False
            adata.var.loc[adata.var.hvg_exclude, 'highly_variable_rank'] = np.nan
            
            L.info("number of hvgs after filtering")
            L.info(adata.var.highly_variable.sum())
            sc.pl.highly_variable_genes(adata,show=False, save ="_exclude_genes_highlyvar.png")
    else:
        sys.exit("exclusion file %s not found, check the path and try again" % args.exclude_file)

if isinstance(mdata, mu.MuData):
    mdata.update()

L.debug(adata.uns['log1p'])
# filter by hvgs
filter_by_hvgs = args.filter_by_hvg

if filter_by_hvgs is True:
    L.info("filtering object to only include highly variable genes")
    mdata.mod["rna"] = adata[:, adata.var.highly_variable]
    if isinstance(mdata, mu.MuData):
        mdata.update()
    adata = mdata.mod["rna"]
    genes = adata.var
    genes['gene_name'] = adata.var.index
    genes.to_csv("filtered_genes.tsv", sep="\t", index=True)
 

adata.layers['logged_counts'] = adata.X.copy()
L.debug(adata.uns['log1p'])
# regress out
if args.regress_out is not None:
    regress_opts = args.regress_out.split(",")
    L.info("regressing out %s" % regress_opts)
    sc.pp.regress_out(adata, regress_opts)


if args.scale is True and args.scale_max_value is None:
    L.info("scaling data with default parameters")
    sc.pp.scale(adata)
elif args.scale is True:
    L.info("scaling data to max value %i" % int(args.scale_max_value))
    sc.pp.scale(adata, max_value=int(args.scale_max_value))
else:
    L.info("not scaling data")
    pass

L.debug(adata.uns['log1p'])
# run pca
L.info("running pca")

if adata.var.shape[0] < int(args.n_pcs):
    L.info("You have less features than number of PCs you intend to calculate")
    n_pcs = adata.var.shape[0] - 1
    L.info("Setting n PCS to %i" % int(n_pcs))    
else:
    n_pcs = int(args.n_pcs)
sc.tl.pca(adata, n_comps=n_pcs, 
                svd_solver=args.pca_solver, 
                random_state=0) 

# do some plots!
sc.pl.pca_variance_ratio(adata, log=True, n_pcs=n_pcs, save=".png")

col_variables = args.color_by.split(",")
col_variables = [a.strip() for a in col_variables]

col_use = [var for var in col_variables if var in adata.obs.columns]
sc.pl.pca(adata, color=col_use, save = "_vars.png")
sc.pl.pca_loadings(adata, components="1,2,3,4,5,6", save = ".png")
sc.pl.pca_overview(adata, save = ".png")


mdata.update()
# save the scaled adata object

mdata.write(args.output_scaled_mudata)
L.debug(adata.uns['log1p'])

L.info("done")



import muon as mu
import scanpy as sc
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import argparse
import os
import itertools
import re

from panpipes.funcs.io import read_yaml
from panpipes.funcs.plotting import get_layer, plot_scatters


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("test logging message")


# parse arguments
parser = argparse.ArgumentParser()
parser.add_argument('--mdata_object',
                    default='',
                    help='')
parser.add_argument('--layers_dict',
                    default='',
                    help='')
parser.add_argument('--scatters_csv',
                    default='',
                    help='')

parser.set_defaults(verbose=True)
args, opt = parser.parse_known_args()

L.info(args)

sc.settings.figdir  = "./scatters/"
sc.set_figure_params(fontsize=12)


mdata = mu.read(args.mdata_object)

layers = read_yaml(args.layers_dict)


df = pd.read_csv(args.scatters_csv)
L.debug(df)

layers = df.applymap(lambda x: get_layer(x,  mdata=mdata, layers=layers))
L.info(layers)

for idx in range(df.shape[0]):
    features_list = df.iloc[idx,:].tolist()
    # convert NAs and X to None
    features_list = [x  if pd.notna(x) else None for x in features_list]
    
    layers_list = layers.iloc[idx,:].tolist()
    # exclude any Nones to get the layers list to be the same length as features
    layers_list = [x for idx, x in enumerate(layers_list) if features_list[idx] is not None]
    
    fname = os.path.join("scatters", "_".join([x for x in features_list if x is not None]) + ".png")
    plot_scatters(mdata, features_list, layers_list)
    # plt.legend(bbox_to_anchor=(1.04,0.5), loc="center left", borderaxespad=0)
    plt.subplots_adjust(right=0.85)
    plt.savefig(fname, bbox_inches="tight")

L.info("Done")




import multiprocessing 

import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import scvi
import argparse
import os
import gc
import muon as mu

from panpipes.funcs.io import read_anndata, read_yaml
from panpipes.funcs.scmethods import run_neighbors_method_choice, X_is_raw
from panpipes.funcs.processing import intersect_obs_by_mod

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--scaled_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--raw_anndata',
                    default=None,
                    help='')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_totalvi.csv',
                    help='')
parser.add_argument('--integration_col_categorical', default=None,
                    help='')
parser.add_argument('--integration_col_continuous', default=None,
                    help='')
parser.add_argument('--figdir', default='./figures',
                    help='')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs", default=50)
parser.add_argument('--neighbors_method', default="scanpy",
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k', default=30,
                    help="neighbors k")
parser.add_argument('--neighbors_metric',default="euclidean",
                    help="neighbor metric, e.g. euclidean or cosine")

args, opt = parser.parse_known_args()
L.info(args)
# scanpy settings
sc.set_figure_params(facecolor="white")
sc.settings.autoshow = False
sc.settings.figdir = args.figdir


# load parameters

threads_available = multiprocessing.cpu_count()

params = read_yaml("pipeline.yml")
params['sample_prefix']



test_script=False
if test_script:
    L.info("this is a test run")
    params['multimodal']['totalvi']['training_args']['max_epochs'] = 10

# ------------------------------------------------------------------
L.info("Running totalvi script")

mdata = mu.read(args.scaled_anndata)

# we need to intersect the prot and rna modalities.
intersect_obs_by_mod(mdata, ['rna', 'prot'])

# this is a copy so we can subset vars and obs without changing the original object
rna = mdata['rna'].copy()
prot = mdata['prot'].copy()

kwargs={}
# in case of more than 1 variable, create a fake column with combined information
if args.integration_col_categorical is not None :
    columns = [x.strip() for x in args.integration_col_categorical.split(",")]
    columns =[ x.replace("rna:","") for x in columns]
    if len(columns) > 1:
        L.info("using 2 columns to integrate on more variables")
        # bc_batch = "_".join(columns)
        rna.obs["bc_batch"] = rna.obs[columns].apply(lambda x: '|'.join(x), axis=1)
        # make sure that batch is a categorical
        rna.obs["bc_batch"] = rna.obs["bc_batch"].astype("category")
    else:
        rna.obs['bc_batch'] = rna.obs[columns] #since it's one
        rna.obs["bc_batch"] = rna.obs["bc_batch"].astype("category")
    batch_categories = list(rna.obs['bc_batch'].unique())
    kwargs["batch_key"] = "bc_batch"
else:
    batch_categories = None
if args.integration_col_continuous is not None :
        rna.obs['bc_batch_continuous'] = rna.obs[args.integration_col_continuous]
        kwargs["continuous_covariate_keys"] = ["bc_batch_continuous"]

L.debug(kwargs)

# add in raw counts as a layer 
# try to find or load the raw counts
if "counts" in rna.layers.keys():
    L.info("raw counts found in layer")
elif "raw_counts" in rna.layers.keys():
    rna.layers["counts"] = rna.layers["raw_counts"].copy() 
    L.info("raw counts found in layer")
elif X_is_raw(rna):
    # this means the X layer is already raw and we can make the layer we need
    L.info("raw counts found in X")
    rna.layers["counts"] = rna.X.copy()
else:
    L.info("merge in raw counts")
    sc_raw = read_anndata(args.raw_anndata, use_muon=True, modality="rna")
    #filter by barcodes in the scaled object
    sc_raw = sc_raw[sc_raw.obs_names.isin(rna.obs_names),: ]
    rna.layers["counts"] = sc_raw.X.copy()

# filter prot outliers 
if params['multimodal']['totalvi']['filter_prot_outliers']:
    # for this to work the user needs to (manually) make a column called prot_outliers
    # actually there is a thing in the qc pipe that calculates outliers, I don't like it very much though
    if "prot_outliers" in mdata['prot'].columns:
        mu.pp.filter_obs(mdata, "prot_outliers")
    else:
        raise ValueError("prot_outliers column not found in mdata['prot'].obs")

# exluding isotypes
if 'isotype' in prot.var.columns:
    prot = prot[:, ~prot.var.isotype]

# filter out mitochondria
if params['multimodal']['totalvi']['exclude_mt_genes']:
    L.info("excluding mt genes")
    rna = rna[:, ~rna.var[params['multimodal']['totalvi']['mt_column']]]
    
# filter by Hvgs

if params['multimodal']['totalvi']['filter_by_hvg']:
    L.info("filtering by highly_variable")
    rna = rna[:, rna.var.highly_variable]

mdata.update()

#make protein obsm pandas
# need to find the raw counts in prot
if X_is_raw(prot):
    X_array = prot.X.copy()
elif "raw_counts" in prot.layers.keys(): 
    X_array = prot.layers['raw_counts'].copy()
else:
    raise AttributeError("raw counts not found in prot, \
                        store in either X or in 'raw_counts' layer")


X_df = pd.DataFrame(X_array.todense(), index=prot.obs_names, columns=prot.var_names)

if X_df.shape[0] == rna.X.shape[0]:
    L.info("adding protein_expression to obsm")
    #check the obs are in the correct order
    X_df = X_df.loc[rna.obs_names,:]
    X_df = X_df.astype('int')
    rna.obsm['protein_expression'] = X_df 
else:
    L.error("dimensions do not match, cannot integrate protein")

# clear up to preserve ram
del X_array, X_df
gc.collect()

mdata.update()

L.debug(rna.obs.columns)
scvi.model.TOTALVI.setup_anndata(
    rna,
    layer="counts",
    protein_expression_obsm_key="protein_expression",
    # categorical_covariate_keys = "bc_batch"
    **kwargs
)

if params['multimodal']['totalvi']['model_args'] is None:
    totalvi_model_args =  {}
else:
    totalvi_model_args =  {k: v for k, v in params['multimodal']['totalvi']['model_args'].items() if v is not None}

print(totalvi_model_args)

if params['multimodal']['totalvi']['training_args'] is None:
    totalvi_training_args = {}
else:
    totalvi_training_args =  {k: v for k, v in params['multimodal']['totalvi']['training_args'].items() if v is not None}

print(totalvi_training_args)

if params['multimodal']['totalvi']['training_plan'] is None:
    totalvi_training_plan = {}
else:
    totalvi_training_plan =  {k: v for k, v in params['multimodal']['totalvi']['training_plan'].items() if v is not None}

print(totalvi_training_plan)


vae = scvi.model.TOTALVI(rna, **totalvi_model_args)

vae.train(**totalvi_training_args, plan_kwargs=totalvi_training_plan)

vae.save(os.path.join("batch_correction", "totalvi_model"), 
                  anndata=False, overwrite=True )


plt.plot(vae.history["elbo_train"], label="train")
plt.plot(vae.history["elbo_validation"], label="validation")
plt.title("Negative ELBO over training epochs")
# plt.ylim(1200, 1400)
plt.legend()
plt.savefig(os.path.join(args.figdir, "totalvi_elbo_plot.png"))

# scvi.model..view_anndata_setup(vae.adata)

# we want to put things back in the original mdata

L.info("""We support the use of mudata as a general framework for multimodal data
        For this reason, the object we save is not the classical anndata
        you would find in scvitools tutorial.
        We chose to save the learned SC representation in the 
        Mudata obsm slot, and any single modality processing in
        its own modality slot
            """)
mdata.obsm["X_totalVI"] = vae.get_latent_representation()

if batch_categories is not None:
    L.debug(batch_categories)
    if type(batch_categories) is not list:
        batch_categories = [batch_categories]
    normX, protein = vae.get_normalized_expression(
        n_samples=25,
        return_mean=True,
        transform_batch=batch_categories
    )

    # this is stored in obsm, because if we have filtered by hvg, it no longer fits specifications for a layer
    mdata['rna'].obsm["totalvi_denoised_rna"] = normX
    # reorder if rna and prot are in different orders
    mdata['prot'].obsm["totalvi_denoised_protein"] = protein.loc[mdata['prot'].obs_names,:]
    #
    df = vae.get_protein_foreground_probability(
        n_samples=25,
        return_mean=True,
        transform_batch=batch_categories
    )
    mdata['prot'].obsm["totalvi_protein_foreground_prob"] = df.loc[mdata['prot'].obs_names,:]

mdata.update()

if int(args.neighbors_n_pcs) > mdata.obsm['X_totalVI'].shape[1]:
    L.warning(f"N PCs is larger than X_totalVI dimensions, reducing n PCs to  {mdata.obsm['X_totalVI'].shape[1] -1}")

n_pcs= min(int(args.neighbors_n_pcs), mdata.obsm['X_totalVI'].shape[1]-1)


# neighbors function has been made to magically calculate mudata neighbors using sc.pp.neighbors
# because we want to use_rep = X_totalvi (as a pseudo single modality) not run multimodal mu.pp.neighbors
run_neighbors_method_choice(mdata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=n_pcs, 
    metric=args.neighbors_metric, 
    use_rep='X_totalVI',
    nthreads=max([threads_available, 6]))

sc.tl.umap(mdata, min_dist=0.4)
sc.tl.leiden(mdata, key_added="leiden_totalVI")

umap = pd.DataFrame(mdata.obsm['X_umap'], mdata['rna'].obs.index)
umap.to_csv(args.output_csv)

mdata.write("tmp/totalvi_scaled_adata.h5mu")

L.info("Done")




import argparse
import pandas as pd
import scanpy as sc

from muon import read 

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
parser.add_argument('--infile',
                    default="./mdata.h5mu",
                    help="mudata object ")
parser.add_argument('--outfile', default='sc_preprocess.txt',
                    help="path for text file output")
args = parser.parse_args()

args, opt = parser.parse_known_args()

# read file
mdata = read(args.infile)
mdata.obs.to_csv(args.outfile, sep='\t')


import pytest
import pandas as pd
import types
# from scpipelines.funcs.processing import extract_parameter_from_fname
import panpipes.funcs as pnp
from panpipes.funcs.io import gen_load_anndata_jobs,update_cellranger_col
from panpipes.pipeline_ingest import  unfilt_file, raw_file
import os
from cgatcore import pipeline as P


PARAMS = P.get_parameters("../scpipelines/pipeline_ingest/pipeline.yml")
PARAMS['submission_file'] = os.path.dirname(__file__) + "/sample_caf.tsv"
PARAMS['sample_prefix'] = "test"

#test generators
def test_gen_load_anndata_jobs():
    caf = pd.read_csv(PARAMS['submission_file'], sep='\t')
    assert isinstance(gen_load_anndata_jobs(caf=caf), types.GeneratorType)
    assert  next(gen_load_anndata_jobs(caf=caf, mode_dictionary={"RNA":True, "ADT":True})) == ('../sample_1_rna/filtered_feature_bc_matrix',
                                                     './tmp/sample1.h5mu', 
                                                     'sample1', 
                                                     'cellranger', 
                                                     '../sample_1_prot/raw_feature_bc_matrix', 'cellranger', 
                                                     None,None, 
                                                     None,None,
                                                     None,None,
                                                     None) 
# all the nones are for repertoire and atac paths 

# test defaults
# this basically checks that the default yml has not changed
def test_filenames():
    assert unfilt_file() == "test_unfilt.h5mu"
    assert raw_file() == "test_raw.h5mu"
    assert update_cellranger_col("path", raw=False) == ("path/filtered_feature_bc_matrix", "cellranger")
    assert update_cellranger_col("path", raw=True) == ("path/raw_feature_bc_matrix", "cellranger")
    # test with temp file
    with open('raw_feature_bc_matrix.h5', 'w'): pass
    assert update_cellranger_col("", raw=True) == ("raw_feature_bc_matrix.h5", "10X_h5")
    os.remove('raw_feature_bc_matrix.h5')




'''
scanpy Normalize script ATAC
'''
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import scipy.io
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import anndata as ad
import os
import argparse
import seaborn as sns
from muon import atac as ac
import muon as mu
import sys
import logging
from panpipes.funcs.scmethods import X_is_raw
from panpipes.funcs.scmethods import findTopFeatures_pseudo_signac
from panpipes.funcs.scmethods import lsi, calc_tech_corr, extract_lsi
from panpipes.funcs.plotting import plot_lsi_corr


L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("testing logger works")

sc.settings.verbosity = 3

parser = argparse.ArgumentParser()

parser.add_argument("--input_mudata",
                    default="mudata_unfilt.h5mu",
                    help="")
parser.add_argument("--output_mudata",
                    default="mudata_unfilt.h5mu",
                    help="")
parser.add_argument("--figdir",
                    default="./figures/atac",
                    help="path to save the figures to")
parser.add_argument("--binarize",
                    default=False,
                    help="whether to binarize the peak counts matrix before any other normalization, like TFIDF")
parser.add_argument("--normalize",
                    default="TFIDF",
                    help="what operation to use to normalize data. Options are: log1p, TFIDF, None")
parser.add_argument("--TFIDF_flavour",
                    default="signac",
                    help="how to log the data. Default to signac(ArchR) default (1) from https://satijalab.org/signac/reference/runtfidf")
parser.add_argument("--min_mean",
                    default=0.05,
                    help="min mean expression for HVF selection")
parser.add_argument("--max_mean",
                    default=1.5,
                    help="max mean expression for HVF selection")
parser.add_argument("--min_disp",
                    default=0.5,
                    help="dispersion min for HVF selection")
parser.add_argument('--n_top_features', default=None,
                    help = "if using scanpy and setting this, selecting these many HVF")
parser.add_argument("--dimred",
                    default="PCA",
                    help="which dimensionality red to use for atac")
parser.add_argument("--n_comps",
                    default=50,
                    help="how many components to compute")
parser.add_argument("--solver",
                    default="arpack",
                    help="what pca solver to use")
parser.add_argument("--dim_remove",
                    default=None,
                    help="which dimensionality red components to remove")
parser.add_argument("--feature_selection_flavour",
                    default=None,
                    help="which HVF selection to perform, 'scanpy' or 'signac'")
parser.add_argument("--min_cutoff",
                    default=None,
                    help="cutoff for Signac's HVF selection")
parser.add_argument("--filter_by_hvf", default=False) 
parser.add_argument("--color_by", default="batch") 


args, opt = parser.parse_known_args()

L.info("running with args:")
L.info(args)

figdir = args.figdir

if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))

mdata = mu.read(args.input_mudata)
atac = mdata.mod['atac']

# save raw counts
if X_is_raw(atac):
    atac.layers["raw_counts"] = atac.X.copy()
elif "raw_counts" in atac.layers :
    L.info("raw_counts layer already exists")
else:
    L.error("X is not raw data and raw_counts layer not found")
    sys.exit("X is not raw data and raw_counts layer not found")


# NORMALIZE
if X_is_raw(atac):
    if args.binarize is True:
        L.info("binarizing peak count matrix")
        ac.pp.binarize(atac)    

    if args.normalize is not None:
        if args.normalize == "log1p":
            if args.binarize:
                L.warning("Careful, you have decided to binarize data but also to normalize per cell and log1p. Not sure this is meaningful")
            sc.pp.normalize_per_cell(atac,counts_per_cell_after=1e4)
            sc.pp.log1p(atac)
            atac.layers["logged_counts"] = atac.X.copy()
        elif args.normalize == "TFIDF":
            if args.TFIDF_flavour=="signac":
                ac.pp.tfidf(atac, scale_factor=1e4, log_tfidf=True, log_tf=False, log_idf=False)
                atac.layers["signac_norm"] = atac.X.copy()
            elif args.TFIDF_flavour=="logTF":
                ac.pp.tfidf(atac, scale_factor=1e4, log_tfidf=False, log_tf=True, log_idf=False)    
                atac.layers["logTF_norm"] = atac.X.copy()
            elif args.TFIDF_flavour=="logIDF":
                ac.pp.tfidf(atac, scale_factor=1e4, log_tfidf=False, log_tf=False, log_idf=True)
                atac.layers["logIDF_norm"] = atac.X.copy()
    else: 
        L.error("We Require a normalization strategy for ATAC")
        sys.exit("Exiting because no normalization was specified. If None was intended, check your pipeline.yml file")


#highly variable feature selection
if "highly_variable" in atac.var: 
    L.warning( "You have %s Highly Variable Features", np.sum(atac.var.highly_variable))
else:

    if args.feature_selection_flavour == "scanpy":
        if args.n_top_features is None:
            sc.pp.highly_variable_genes(atac, min_mean=float(args.min_mean), 
            max_mean=float(args.max_mean), min_disp=float(args.min_disp))
        else:
            sc.pp.highly_variable_genes(atac, n_top_genes=int(args.n_top_features))
    elif args.feature_selection_flavour == "signac":
        findTopFeatures_pseudo_signac(atac, args.min_cutoff)
    else:
        L.warning("No highly variable feature selection was performed!")

if "highly_variable" in atac.var: 
    L.warning( "You have %s Highly Variable Features", np.sum(atac.var.highly_variable))

# filter by hvgs
filter_by_hvgs = args.filter_by_hvf

if filter_by_hvgs:
    L.info("filtering object to only include highly variable features")
    mdata.mod["atac"] = atac[:, atac.var.highly_variable]
    if isinstance(mdata, mu.MuData):
        mdata.update()
    atac = mdata["atac"]
    genes = atac.var
    genes['gene_name'] = atac.var.index
    genes.to_csv("filtered_variable_features.tsv", sep="\t", index=True)

# The combined steps of TF-IDF followed by SVD are known as latent semantic indexing (LSI), 
# and were first introduced for the analysis of scATAC-seq data by Cusanovich et al. 2015.

if args.dimred == "LSI" and args.normalize == "TFIDF":
    lsi(adata=atac, num_components=int(args.n_comps))
else:
    L.info("Applying LSI on logged_counts is not recommended. Changing dimred to PCA")
    args.dimred = "PCA"
if args.dimred == "PCA":
    sc.pp.scale(atac)
    atac.layers["scaled_counts"] = atac.X.copy()
    sc.tl.pca(atac, n_comps=int(args.n_comps), svd_solver=args.solver, 
    random_state=0)

col_variables = args.color_by.split(",")
col_variables = [a.strip() for a in col_variables]
col_use = [var for var in col_variables if var in atac.obs.columns]

#some plotting before removing the components

if args.dimred =='PCA':
    sc.pl.pca_variance_ratio(atac, log=True, n_pcs=int(args.n_comps), save=".png")
    sc.pl.pca(atac, color=col_use, save = "_vars.png")
    sc.pl.pca_loadings(atac, components="1,2,3,4,5,6", save = ".png")
    sc.pl.pca_overview(atac, save = ".png")
if args.dimred =='LSI':
    sc.pl.embedding(atac, color=col_use,basis="X_lsi", save = ".png")
    correlation_df = calc_tech_corr(extract_lsi(atac), 
        tech_covariates = ['n_genes_by_counts', 'total_counts'])

    correlation_df.to_csv(os.path.join(args.figdir,"tech_covariates_corr_LSI.tsv"), sep='\t')

    plot_lsi_corr(correlation_df, tech_covariates=['n_genes_by_counts', 'total_counts'],
        filename=os.path.join(args.figdir,"LSI_corr_plot.png"))


if args.dim_remove is not None:
    L.info("removing component from dimred")
    dimrem=int(args.dim_remove)
    if args.dimred == "LSI":
        atac.obsm['X_lsi'] = atac.obsm['X_lsi'][:,dimrem:]
        atac.varm["LSI"] = atac.varm["LSI"][:,dimrem:]
        atac.uns["lsi"]["stdev"] = atac.uns["lsi"]["stdev"][dimrem:]
    if args.dimred == "PCA":
        atac.obsm['X_pca'] = atac.obsm['X_pca'][:,dimrem:]
        atac.varm["PCs"] = atac.varm["PCs"][:,dimrem:]
        
mdata.update()
mdata.write(args.output_mudata)

L.info("Done")




#! /usr/bin/env python

from ruffus import *
import sys
import os
import re
from cgatcore import pipeline as P
import pandas as pd
import cgatcore.iotools as IOTools
from panpipes.funcs.io import gen_load_spatial_jobs
# from scpipelines.funcs.processing import intersection
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning) 

import warnings

# warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
# warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)

import yaml


PARAMS = P.get_parameters(
    ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
     "pipeline.yml"])

PARAMS['py_path'] =  os.path.join(os.path.dirname(os.path.dirname(__file__)), 'python_scripts')
PARAMS['r_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'R_scripts')
PARAMS['resources_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")
job_kwargs = {}

if PARAMS['condaenv'] is not None:
    job_kwargs["job_condaenv"] =PARAMS['condaenv']

print(PARAMS)
#------------------------------------------------------------------------------------------------
## Create a dictionary of modalities
#------------------------------------------------------------------------------------------------
mode_dictionary = {'spatial': True} #PARAMS["modalities"]
#{'spatial': True}
print(mode_dictionary)
#------------------------------------------------------------------------------------------------
# setup dirs
#------------------------------------------------------------------------------------------------

@originate("logs/setup_dirs.sentinel")
def set_up_dirs(log_file):
    os.makedirs("logs", exist_ok=True)
    os.makedirs("tmp", exist_ok=True)
    os.makedirs("figures", exist_ok=True)
    IOTools.touch_file(log_file)
    pass

# -----------------------------------------------------------------------------------------------
## Creating h5mu from filtered data files
# -----------------------------------------------------------------------------------------------


def gen_load_spatial_anndata_jobs():
    print(PARAMS['submission_file'])
    caf = pd.read_csv(PARAMS['submission_file'], sep='\t')
    print(gen_load_spatial_jobs(caf,mode_dictionary={'spatial': True}))
    global assays
    assays = {}
    return gen_load_spatial_jobs(caf,mode_dictionary={'spatial': True})
   


@follows(mkdir("logs"))
@follows(mkdir("tmp"))
@files(gen_load_spatial_anndata_jobs)
def load_mudatas(spatial_path, outfile, 
                 sample_id,
                 spatial_filetype, 
                 spatial_counts, 
                 spatial_metadata, 
                 spatial_transformation):
    
    path_dict = {'spatial':spatial_path}
                 
    print(path_dict)
    print('sample_id = %s' % str(sample_id))
    print('outfile = %s' % str(outfile))
    print('spatial_filetype = %s' % str(spatial_filetype))
    print('spatial_counts = %s' % str(spatial_counts))
    if spatial_filetype == "vizgen":
        print('spatial_metadata = %s' % str(spatial_metadata))
        print('spatial_transformation = %s' % str(spatial_transformation))
    else:
        print("visium")
    modality_dict = {k:True if path_dict[k] is not None else False for k,v in {'spatial': True}.items() }
    print(modality_dict)

    assays[outfile] = spatial_filetype
    
    cmd = """
        python %(py_path)s/make_mudataspatial_from_csv.py 
        --mode_dictionary "%(modality_dict)s"
        --sample_id %(sample_id)s
        --output_file %(outfile)s 
        --spatial_filetype %(spatial_filetype)s
        --spatial_infile %(spatial_path)s
        --spatial_counts %(spatial_counts)s
    """
    if spatial_filetype == "vizgen":
        cmd += """
        --spatial_metadata %(spatial_metadata)s 
        --spatial_transformation %(spatial_transformation)s
        """
    cmd += " > logs/make_mudatas_%(sample_id)s.log"
    print(cmd)
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)




@follows(load_mudatas)
@follows(mkdir("qc.data"))
@follows(mkdir("./figures"))
@transform(load_mudatas,
           regex("./tmp/(.*)_raw.h5(.*)"), 
           r"./logs/spatialQC_\1.log")
def spatialQC(infile,log_file):
    spatial_filetype = assays[infile]
    resources_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")
    outfile = infile.replace("_raw","_unfilt")
    outfile = outfile.replace("tmp", "qc.data")
    cmd = """
            python %(py_path)s/run_scanpyQC_spatial.py
              --input_anndata %(infile)s
              --spatial_filetype %(spatial_filetype)s
              --outfile %(outfile)s
              --figdir ./figures
              """
    if PARAMS['ccgenes'] is not None:
        if PARAMS['ccgenes'] == "default":
            ccgenesfile = resources_path + "/cell_cycle_genes.tsv"
        else:
            ccgenesfile = PARAMS['ccgenes']
        cmd += " --ccgenes %(ccgenesfile)s"
    if PARAMS['custom_genes_file'] is not None:
        if PARAMS['custom_genes_file'] == "default":
            customgenesfile = resources_path + "/qc_genelist_1.0.csv"
        else:
            customgenesfile = PARAMS['custom_genes_file']
        cmd += " --customgenesfile %(customgenesfile)s"
    if PARAMS['score_genes'] is not None:
        cmd += " --score_genes %(score_genes)s"
    if PARAMS['calc_proportions'] is not None:
        cmd += " --calc_proportions %(calc_proportions)s"
    cmd += " > %(log_file)s"

    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)


def run_plotqc_query(pqc_dict):
    # avoid deleting from PARAMS
    pqc = pqc_dict.copy()
    del pqc['grouping_var']
    return any([x != None for x in pqc.values()])


@follows(spatialQC)
@follows(mkdir("./figures/spatial"))
@active_if(run_plotqc_query(PARAMS['plotqc']))
@transform(load_mudatas, 
           regex("./tmp/(.*)_raw.h5(.*)"),
           r"./logs/qcplot.\1.log")
def plotQC_spatial(unfilt_file,log_file):
    spatial_filetype = assays[unfilt_file]
    unfilt_file = unfilt_file.replace("_raw","_unfilt")
    unfilt_file = unfilt_file.replace("tmp", "qc.data")
    cmd = """
            python %(py_path)s/plot_qc_spatial.py
             --input_mudata %(unfilt_file)s
             --spatial_filetype %(spatial_filetype)s
             --figdir ./figures/spatial
            """
    if PARAMS['plotqc']['grouping_var'] is not None:
        cmd += " --grouping_var %(plotqc_grouping_var)s"
    if PARAMS['plotqc']['spatial_metrics'] is not None:
        cmd += " --spatial_qc_metrics %(plotqc_spatial_metrics)s"
    cmd += " > %(log_file)s "
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)

    
#----- end

@follows(plotQC_spatial)
def full():
    """
    All cgat pipelines should end with a full() function which updates,
    if needed, all branches of the pipeline.
    The @follows statement should ensure that all functions are covered,
    either directly or as prerequisites.
    """
    pass


def main(argv=None):
    if argv is None:
        argv = sys.argv
    P.main(argv)


if __name__ == "__main__":
    sys.exit(P.main(sys.argv))

    


"""
CGAT pipeline for clustering single cell data with Scanpy.
# ASSUMED INPUT: mudata with normalized and scaled data in the relevant layers
# This pipeline is designed to follow on from pipeline_integration.py

"""

from ruffus import *
import sys
import os
from cgatcore import pipeline as P
import pandas as pd
import cgatcore.iotools as IOTools
import re
from itertools import chain, product
import glob

from panpipes.funcs.processing import extract_parameter_from_fname
PARAMS = P.get_parameters(
    ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
     "pipeline.yml"])

PARAMS['py_path'] =  os.path.join(os.path.dirname(os.path.dirname(__file__)), 'python_scripts')
PARAMS['r_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'R_scripts')
PARAMS['mudata_with_knn'] = 'mudata_w_neighbors.h5mu'

job_kwargs={}
if PARAMS['condaenv'] is not None:
    job_kwargs["job_condaenv"] =PARAMS['condaenv']


@originate("logs/setup_dirs.sentinel")
def set_up_dirs(log_file):
    os.mkdir("logs")
    mods =  [key for key, value in PARAMS['modalities'].items() if value is True]
    if PARAMS["multimodal"]["run_clustering"] is True :
        mods.append("multimodal")
    for mod in mods:
        os.makedirs(os.path.join(mod, "figures"))
    IOTools.touch_file(log_file)
    pass
## ------------------------------------
## Single modality scripts
## ------------------------------------

# -----------------------------------=
# neighbors
# --------------------------------------
@follows(set_up_dirs)
@originate(PARAMS['mudata_with_knn'])
def run_neighbors(outfile):
    mods =  [key for key, value in PARAMS['modalities'].items() if value is True]
    if  any([PARAMS['neighbors'][mod]['use_existing'] is False for mod in mods]):
        # this means we want to rerun neighbors for at least one assay
        #we want to replace thhe scaled obj with the new neighbors
        log_file="logs/run_single_mod_neighbors.log"
        cmd="""
        python %(py_path)s/rerun_find_neighbors_for_clustering.py \
            --infile %(scaled_obj)s \
            --outfile %(outfile)s  \
            --neighbor_dict '%(neighbors)s' \
            --n_threads %(resources_threads_high)s
            """
        cmd += " > %(log_file)s"
        job_kwargs["job_threads"] = PARAMS['resources_threads_high']
        P.run(cmd, **job_kwargs)
    else:
        P.run('ln -s %(scaled_obj)s %(outfile)s', without_cluster=True)




# ------------------------------------
# UMAP
# ------------------------------------

def gen_umap_jobs():
    """
    Generate find neighbor jobs with all parameter combinations.
    """
    # same infile for all jobs
    # define files based on jobs
    infile = PARAMS['mudata_with_knn']
    mods =  [key for key, value in PARAMS['modalities'].items() if value is True]
    if PARAMS["multimodal"]["run_clustering"] is True :
        mods.append("multimodal")
    for mod in mods:
        for md in PARAMS['umap'][mod]['mindist']:
            if PARAMS['umap'][mod]['mindist'] is not None:
                output_file = os.path.join(mod,  'md' + str(md) +"_umap.txt.gz")
                log_file = os.path.join("logs","_".join([mod,  'md' + str(md) +"_umap.log"]))
                yield [infile, output_file, mod, md, log_file]

@follows(run_neighbors)
@follows(set_up_dirs)
@files(gen_umap_jobs)
def calc_sm_umaps(infile, outfile, mod, mindist, log_file):
    prefix = os.path.split(infile)[0]
    cmd = """
        python %(py_path)s/run_umap.py \
            --infile %(infile)s \
            --outfile %(outfile)s \
            --min_dist %(mindist)s 
            """
    if mod is not None and mod != "multimodal":
        # if this is not specified it will use rna default
        cmd += " --modality %(mod)s"
    elif mod=="multimodal":
        if PARAMS['multimodal_integration_method'].lower() == "wnn":
            cmd += " --neighbors_key wnn"
    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)


# ------------------------------------
# Clustering
# ------------------------------------

def gen_cluster_jobs():
    """
    Generate find neighbor jobs with all parameter combinations.
    """
    # same infile for all jobs
    # define files based on jobs
    infile = PARAMS['mudata_with_knn']
    mods =  [key for key, value in PARAMS['modalities'].items() if value is True]
    if PARAMS['multimodal']['run_clustering'] is True:
        mods.append("multimodal")
    for mod in mods:
        for res in PARAMS['clusterspecs'][mod]['resolutions']:
            if PARAMS['clusterspecs'][mod]['resolutions'] is not None:
                alg = PARAMS['clusterspecs'][mod]['algorithm']
                output_file = os.path.join(mod, 'alg' + alg + '_res' + str(res), "clusters.txt.gz")
                log_file = os.path.join("logs", "_".join([mod, 'alg' + alg + '_res' + str(res), "clusters.log"]))
                yield [infile, output_file, mod, res, alg, log_file]

@follows(set_up_dirs)
@files(gen_cluster_jobs)
@follows(run_neighbors)
def calc_cluster(infile, outfile,  mod, res, alg, log_file):
    cmd = """python %(py_path)s/run_clustering.py 
            --infile %(infile)s 
            --outfile %(outfile)s 
            --resolution %(res)s 
            --algorithm %(alg)s
    """ 
    if mod is not None and mod != "multimodal":
        # if this is not specified it will use rna default
        cmd += " --modality %(mod)s"
    elif mod=="multimodal":
        if PARAMS['multimodal_integration_method'].lower() == "wnn":
            cmd += " --neighbors_key wnn"
    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)


@collate(calc_cluster,
         regex("(.*)/(.*)/clusters.txt.gz"),
         r"\1/all_res_clusters_list.txt.gz")
def aggregate_clusters(infiles, outfile):
    print(infiles)
    print(outfile)
    infiles_str = ','.join(infiles)
    cmd = "python %(py_path)s/aggregate_csvs.py \
               --input_files_str %(infiles_str)s \
               --output_file %(outfile)s \
               --clusters_or_markers clusters > logs/aggregate_clusters.log"
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']           
    P.run(cmd, **job_kwargs)


@collate([[calc_cluster], [calc_sm_umaps]], formatter(), PARAMS['sample_prefix'] + "_clustered.h5mu")
def collate_mdata(infiles,outfile):
    cluster_files = infiles[0]
    cluster_mods = [os.path.dirname(os.path.dirname(x)) for x in cluster_files]
    cluster_col = [os.path.basename(os.path.dirname(x)) for x in cluster_files]
    cluster_col = [re.sub("alg", "", x) for x in cluster_col]
    cluster_df = pd.DataFrame(zip(cluster_mods,cluster_col,cluster_files), columns = ['mod', 'new_key', 'fpath'])
    cluster_df.to_csv("cluster_file_paths.csv", index=None)
    umap_files = infiles[1]
    print(umap_files)
    umap_mods = [os.path.dirname(x) for x in umap_files]
    umap_key = [extract_parameter_from_fname(x, "md", prefix="") for x in umap_files]
    umap_key = ["X_umap_mindist_" + str(x) for x in umap_key]
    umap_df = pd.DataFrame(zip(umap_mods,umap_key,umap_files), columns = ['mod', 'new_key', 'fpath'])
    umap_df.to_csv("umap_file_paths.csv", index=None)
    
    cmd = """python %(py_path)s/collate_mdata.py
        --clusters_files_csv cluster_file_paths.csv
        --umap_files_csv umap_file_paths.csv
        --output_mudata %(outfile)s 
    """
    if PARAMS['full_obj'] is None:
        mdata_in = PARAMS['mudata_with_knn']
        cmd += "--input_mudata %(mdata_in)s"
    else:
        cmd += "--input_mudata  %(full_obj)s"
    cmd += " > logs/collate_data.log"
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)


@transform(collate_mdata, 
            formatter(""),
            'logs/plot_clusters_umaps.log')
def plot_cluster_umaps(infile, log_file,):
    # get associated umap
    mods =  [key for key, value in PARAMS['modalities'].items() if value is True]
    if PARAMS['multimodal']['run_clustering']:
        mods.append("multimodal")
    mods = ','.join(mods)
    cmd = """python %(py_path)s/plot_cluster_umaps.py \
    --infile %(infile)s 
    --modalities '%(mods)s'
    """
    cmd += " >> %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, jobs_limit=1, **job_kwargs)

@transform(aggregate_clusters, regex("(.*)/all_res_clusters_list.txt.gz"),
            r'logs/\1_clustree.log',
            r'\1/figures/clustree.png', ) 
def plot_clustree(infile, log_file, outfile):
    # convert infiles to comma sep. string
    prefix = re.sub('_dir', '', os.path.dirname(infile))
    # call R
    cmd = "Rscript %(r_path)s/plotclustree.R \
        --infile %(infile)s  \
        --plot_title %(prefix)s \
        --outfile %(outfile)s > %(log_file)s"
    
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd,  **job_kwargs)



# # all the defs
# # def clustering_umbrella
@follows(calc_cluster, collate_mdata, plot_cluster_umaps, plot_clustree)
@originate("logs/cluster_analysis.sentinel")
def cluster_analysis(fname):
    IOTools.touch_file(fname)
    pass


# ------------------------------------
# Markers
# ------------------------------------

@subdivide(calc_cluster, regex("(.*)/(.*)/clusters.txt.gz"),
           r"\1/\2/*_markers.txt",
           r"\1/\2")
def gen_marker_jobs(infile, outfile, base_dir):
    mods =  [key for key, value in PARAMS['modalities'].items() if value is True]
    for md in mods:
        IOTools.touch_file(os.path.join(base_dir, md + "_markers.txt"))

@follows(collate_mdata)
@transform(gen_marker_jobs,
           regex("(.*)/(.*)/(.*)_markers.txt"),
           r"logs/\1_\2_\3_markers.log",
           r"\1/\2/\3_markers",
           r"\1",
           r"\1/\2",
           r"\3"
           )
def find_markers(infile, log_file, outfile_prefix, base_mod, cluster_dir, data_mod):
    """
    Runs scanpy.tl.rank_gene_groups in parallel for each cluster
    """
    anndata_file=PARAMS['sample_prefix'] + "_clustered.h5mu" + "/" + data_mod
    cluster_file = os.path.join(cluster_dir, "clusters.txt.gz")
    cmd = """
    python %(py_path)s/run_find_markers_multi.py 
    --infile %(anndata_file)s
    --cluster_file %(cluster_file)s 
    --output_file_prefix %(outfile_prefix)s 
    --mod %(data_mod)s
    """
    min_cells = PARAMS["markerspecs"][data_mod]["mincells"]
    cmd += " --mincells %(min_cells)s"
    layer_choice = PARAMS["markerspecs"][data_mod]["layer"]
    cmd += " --layer '%(layer_choice)s'"
    testuse = PARAMS["markerspecs"][data_mod]["method"]
    cmd += " --testuse '%(testuse)s'"
    if PARAMS['markerspecs'][data_mod]['pseudo_seurat'] is True:
        min_pct =  PARAMS["markerspecs"][data_mod]["minpct"]
        threshuse =  PARAMS["markerspecs"][data_mod]["threshuse"]
        cmd += """ --pseudo_seurat True 
        --minpct %(min_pct)s 
        --threshuse %(threshuse)s 
        """
    cmd += " > %(log_file)s "
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)




# # limit jobs because reading the h5 file simultaneouly is problematic
# @collate([ [collate_mdata],[find_markers]],
#         formatter(),
#         os.path.join("logs", "marker_heatmap.log"))
# def heatmap_markers(infiles, log_file):
#     """
#     Plots a standard Seurat Heatmap, or a Complex Heatmap
#     """
#     # print(infiles, log_file)
#     marker_files = infiles[1]
#     scaled_obj = infiles[0][0]
#     cmd = """
#     Rscript %(r_path)s/marker_heatmaps.R \
#     --anndata %(scaled_obj)s \
#     --marker_files %(marker_file)s \
#     --docomplex %(plotspecs_docomplex)s \
#     --subgroup %(plotspecs_discrete_variables)s > %(log_file)s"""
#     P.run(cmd, job_threads=PARAMS['resources_threads_high'])

# this transforms gen_marker_jobs instead of find_markers because gen_marker_jobs creates empty marker files
# which are then filled by find_markers.
@follows(collate_mdata)
@follows(find_markers)
@transform(gen_marker_jobs,
            regex(r"(.*)/alg(.*)/(.*)_markers.txt"),
            r"logs/cluster\1_alg\2_exprs_\3_dotplots.log",
            r"\1/alg\2/figures/dotplot_top_markers_\3.png",
            r"\1/alg\2/figures",
             r"\1", r"\2", r"\3")
def plot_marker_dotplots(marker_file, log_file, outfile, 
                         fig_path, cluster_mod, cluster_col, expr_mod):
    """
    Plots some additional marker plots
    Read in the h5mu file, pull the cluster col from the h5mu obs.
    then pull the epxression values from the expr_mod.
    
    """
    data_obj=PARAMS['sample_prefix'] + "_clustered.h5mu"
    # check there is a figures directory
    if not os.path.exists(fig_path):
        os.makedirs(fig_path)
    cmd = """
        python %(py_path)s/plot_scanpy_markers.py \
            --infile %(data_obj)s \
            --modality %(expr_mod)s
            --marker_file %(marker_file)s \
            --figure_prefix %(fig_path)s \
            --n %(plotspecs_top_n_markers)s
            
     """
    if cluster_mod != "multimodal":
        cluster_col = cluster_mod + ":" + cluster_col
    cmd += " --group_col %(cluster_col)s"
    layer_choice = PARAMS["markerspecs"][expr_mod]["layer"]
    cmd += " --layer %(layer_choice)s"
    cmd += " > %(log_file)s "
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)





@follows(find_markers, plot_marker_dotplots)
@originate("marker_analysis.sentinel")
def marker_analysis(fname):
    IOTools.touch_file(fname)
    pass





# #--------
# # Generate cellxgene
# #--------

# cellxgene_file = PARAMS['sample_prefix'] + "_" + PARAMS['modality'] + "_cellxgene.h5ad"
# # @originate(cellxgene_file, add_inputs(gen_neighbor_jobs))
# @merge(calc_neighbors, cellxgene_file)
# def cellxgene(infiles, cellxgene_file):
#     print(infiles)
#     print(cellxgene_file)
#     desired_match = "nneigh" + str(PARAMS['cellxgene_nneighbours']) + "_pcs" + str(PARAMS['cellxgene_npcs'])
#     print(desired_match)
#     desired_match = re.compile(".*" + desired_match + ".*")
#     infiles = list(filter(desired_match.match, infiles))
#     print(infiles)
#     if len(infiles) > 1:
#         sys.exit("too many matches, ask charlotte to debug")
#     elif len(infiles) == 0:
#         sys.exit("no matches, did you fill in the cellxgene section of the yml?")
#     else:
#         infile = infiles[0]
    
#     umaps = glob.glob(os.path.dirname(infile) +"/*_umap.txt.gz")
#     umaps_str = ",".join(umaps)

#     clusters_file = os.path.dirname(infile) + "/all_res_clusters_list.txt.gz"
#     best_cluster_res = PARAMS['clusterspecs_algorithm'] + "_res_" + str(PARAMS['cellxgene_cluster_res'])
#     cmd="""
#     python  %(py_path)s/generate_cellxgene.py
#     --input_anndata %(infile)s \
#     --output_anndata %(cellxgene_file)s \
#     --sample_prefix %(sample_prefix)s
#     --clusters %(clusters_file)s \
#     --umaps %(umaps_str)s \
#     --best_md %(cellxgene_umap_md)s \
#     --best_cluster_col %(best_cluster_res)s
#     """
#     if PARAMS['use_muon']:
#         cmd += " --use_muon True"
#     if PARAMS['modality']:
#         # if this is not specified it will use rna default
#         cmd += " --modality %(modality)s"
#     print(cmd)
#     P.run(cmd, job_threads=PARAMS['resources_threads_medium'])

# #--------


@follows(cluster_analysis, marker_analysis )
@originate("cleanup_done.txt")
def cleanup(file):
    # remove any ctmp fails
    P.run("rm ctmp*", without_cluster=True)
    # delete empty dirs
    P.run("find ./ -empty -type d -delete", without_cluster=True)


@follows(cleanup)
def full():
    """
    All cgat pipelines should end with a full() function which updates,
    if needed, all branches of the pipeline.
    The @follows statement should ensure that all functions are covered,
    either directly or as prerequisites.
    """
    pass


def main(argv=None):
    if argv is None:
        argv = sys.argv
    P.main(argv)


if __name__ == "__main__":
    sys.exit(P.main(sys.argv))



import argparse
import numpy as np
import pandas as pd
import harmonypy as hm
import seaborn as sns
import os
from panpipes.funcs.io import read_yaml
import matplotlib.pyplot as plt

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter("%(asctime)s: %(levelname)s - %(message)s")
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--combined_umaps_df")
parser.add_argument("--cell_meta_df")
parser.add_argument("--integration_dict")
parser.add_argument("--fig_dir")
args = parser.parse_args()
L.info(args)

# load combined umaps and cellmtd file
# load metadata
cell_meta_df = pd.read_csv(args.cell_meta_df, index_col=0)
batch_dict = read_yaml(args.integration_dict)
umaps = pd.read_csv(args.combined_umaps_df, sep="\t", index_col=0)
L.info("fixing the batch_keys dictionary")
for k in batch_dict.keys():
    v=batch_dict[k]
    if k =="multimodal":
        mk = []
        exc = []
        for v2 in v:
            if v2 in cell_meta_df.columns:
                mk.append(v2)
            else:
                exc.append(v2)
                for k in batch_dict.keys():
                    if k!="multimodal":
                       tmp=[k + ":" +e for e in exc] #add the prepending modality
                       for t in tmp:
                           if t in cell_meta_df.columns:
                               mk.append(t)
        v = mk
        batch_dict[k] = v
   # else:
   #     v =[k + ":" +v1 if k!="multimodal" else v1 for v1 in v  ] #add the prepending modality
   #     batch_dict[k] = v
    if len(v) > 1:
        mod_meta_df= cell_meta_df[v].dropna()
        cell_meta_df[str(k)+ ":bc_batch"] = mod_meta_df.apply(lambda x: "|".join(str(x)), axis=1)
        batch_dict[k].append(k+ ":bc_batch")
    L.info("batch keys %s" %(batch_dict[k]))


for md in batch_dict.keys():
    L.info("Running lisi on modality: %s" % md)
    # get one df per method for this modality
    splits = dict(list(umaps[umaps["mod"]==md].groupby("method")))
    lisi_results = []
    columns = batch_dict[md]
    for k, xx in splits.items():
        L.info("computing lisi for correction %s" % k)
        # compute LISI for each batch
        umap_coords = xx.loc[:,["umap_1", "umap_2"]].to_numpy()
        # check it"s in the correct order
        batch_df = cell_meta_df.loc[xx.index,:]
        res = hm.compute_lisi(umap_coords, batch_df, columns)
        lisi_results.append(pd.DataFrame(res, index=batch_df.index, columns=columns))
    # put LISI scores into a pandas df
    lisi_df = pd.concat(lisi_results, keys=splits.keys()).reset_index().rename(columns={"level_0":"method", "level_1":"cellbarcode", "index":"cellbarcode"})
    lisi_df.to_csv(os.path.join(args.fig_dir, md, "LISI_scores.csv"), index=False)
    # # make a density  plot of LISI scores.
    L.info("plotting lisi density")
    plot_df = lisi_df.melt(id_vars=['cellbarcode', 'method'], var_name="integration_variable", value_name="LISI score")
    plot_df["integration_variable"] = plot_df["integration_variable"].astype("category")
    plot_df = plot_df.rename(columns={"method": "Correction"})
    plot_df["Correction"] = plot_df["Correction"].astype('category')
    # # reorder categories to match other plots
    methods = plot_df["Correction"].unique().tolist()
    if "none" in methods:
        methods.insert(0, methods.pop(methods.index("none")))
        plot_df["Correction"] = plot_df["Correction"].cat.reorder_categories(methods)
    # do plot
    fig, ax = plt.subplots(nrows=1, ncols=len(columns), figsize=(len(columns)*4, 4))
    if len(columns) > 1:
        for idx, col in enumerate(columns):
            sns.kdeplot(data=plot_df[plot_df.integration_variable==col],
                x="LISI score", hue="Correction", ax=ax[idx])
            ax[idx].set_title("integrated by :" + col) 
    else:
        sns.kdeplot(data=plot_df,
            x="LISI score", hue="Correction", ax=ax)
        ax.set_title("integrated by :" + columns[0])  
    fig.savefig(os.path.join(args.fig_dir, md, "LISI_scores.png"))
    plt.clf()



L.info("Done")